#!@TCLSH@
# This is a -*- Tcl -*- script.

# rcsmv --- Moving files with RCS counterparts

# Copyright (C) 1996, 1998 Stefan Hornburg

# Author: Stefan Hornburg <racke@gundel.han.de>
# Maintainer: Stefan Hornburg <racke@gundel.han.de>
# Version: see below

# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.

# This file is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this file; see the file COPYING.  If not, write to the Free
# Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

set auto_path [concat "@jultafdir@" $auto_path]
package require Error
package require File
package require Flow 0.0.6
package require Sequence
package require Shell 0.0.9
package require Getopt

# ---------------------------
# setup option specifications
# ---------------------------

set optspecs {{"help" do_help "display this help and exit"}}

# ------------------------------------------
# `usage' displays simple usage instructions
# ------------------------------------------

proc usage {} {
	puts stdout {Usage: rcsmv [OPTION]... SOURCE DEST}
	puts stdout {   or: rcsmv [OPTION]... SOURCE... DIRECTORY}
	puts stdout ""
	puts stdout {Options:}
}

# -----------------------------------------
# strip options from command line arguments
# -----------------------------------------

eval Juf::Getopt::get_options $optspecs

if $do_help {
	# print usage instructions
	usage
	eval Juf::Getopt::print $optspecs
	exit 0
}

# -------------------------
# check number of arguments
# -------------------------

if {[llength $argv] == 0} {
	Juf::Error::fatal "missing file arguments"
}

if {[llength $argv] == 1} {
	Juf::Error::fatal "missing file argument"
}

set dest [Juf::Sequence::pop argv]
if {[llength $argv] > 1} {
	if {![file exists $dest] || ![file isdir $dest]} {
		Juf::Error::fatal \
				"when moving multiple files, last argument must be a directory"
	}
} 

# -----------------
# create namespaces
# -----------------

namespace eval Juf {namespace eval File {namespace eval RCS {}}}}

# -------------------------------------------------------------
# COMMAND: name FILE
#
# Returns RCS pathname for working file FILE or an empty string
# if FILE is not under version control.
# -------------------------------------------------------------

proc Juf::File::RCS::pathname {file} {
	if {[Juf::Shell::run rlog -R $file] == 0} {
		set Juf::Shell::output
	}
}

# main loop
while {[llength $argv]} {
	set source [Juf::Sequence::shift argv]
	# check first if source file and its RCS file exists
	if [file exists $source] {
		if ![string length [set vcfile [Juf::File::RCS::pathname $source]]] {
			Juf::Error::error $source "File not under version control"
			continue
		}
	} else {
		Juf::Error::error $source "No such file or directory"
		continue
	}
	continue
	# move working file to new location
	if [catch "file rename {$source} {$dest}" result] {
		Juf::Error::error $source $result
		continue
	}
}

Juf::Error::finish

# rcsmv - move RCS file
#
# in:	$name		source file name 
#		$target	    target file name
#
# out:	1 if successful, 0 otherwise

proc rcsmv {name target} {
	set srcdir [file dirname $name]
	set srcfile [file tail $name] 
	set tdir [file dirname $target]
	set tfile [file tail $target]

	if {[wtaf_shell_exec "rlog -R $name" rcsname]} {
		set rcsname [string trimright $rcsname "\n"]

		# determine additional directory part
		# -----------------------------------

		if [jultaf_file_iscwd $srcdir] {set origlen 0} else {
			set origlen [string length $srcdir] 
		}
		set rcsdir [string range [file dirname $rcsname] $origlen e]

		# determine additional file name part
		# -----------------------------------
		
		set origlen [string length $srcfile]
		set rcsx [string range [file tail $rcsname] $origlen e]

		# some sanity checks
		# ------------------

		set rcschk [file join [file dirname $rcsname] [file tail $rcsname]]
		set mychk [file join $srcdir $rcsdir "$srcfile$rcsx"]
		wtaf_dbg_out rcsmv_flags_rcschk $rcschk
		wtaf_dbg_out rcsmv_flags_rcschk $mychk
		if {($rcsx == "" && $rcsdir == "") || ($rcschk != $mychk && [file join $srcdir $rcschk] != $mychk)} {
			wtaf_err_err "$rcsname" "suspicious output of `rlog -R $name'"
			return 0
		}
	} else {
		puts stderr $rcsname
		return 0
	}

	set newloc [file join $tdir $rcsdir "$tfile$rcsx"]

	if [wtaf_shell_exec "mv -f $rcsname $newloc" result] {
		return 1
	}

	puts stderr $result
	return 0
}


if {[llength $argv] == 2} {
	set first [lindex $argv 0]
	set second [lindex $argv 1]

	if {[file isdir $first]} {
		# ----------------
		# Rename directory
		# ----------------
		if {[file exists $second]} {
			if {! [file isdir $second]} {
				wtaf_err_fat "cannot move `$first' to `$second'"\
						"Not a directory"
			}
		}
		if {! [wtaf_shell_exec "mv -f $first $second" result]} {
			puts stderr $result
			exit 1
		}
		exit 0
	}

	if {! [file isdir $second]} {
		# -----------------------------------
		# Rename file and associated RCS file
		# -----------------------------------

		if {[wtaf_shell_exec "mv -f $first $second" result]} {
			if [rcsmv $first $second] { exit 0 } else { exit 1}
		}
	}
}

set rcsmv_targetdir [lindex $argv [expr $argc - 1]]
if {! [file isdir $rcsmv_targetdir]} {
	wtaf_err_fat $rcsmv_targetdir\
			"when moving multiple files, last argument must be a directory"
}

foreach file [lrange $argv 0 [expr $argc - 2]] {
	wtaf_dbg_out rcsmv_flags_source "Source file: $file"

	# Check if file exists
	if [file exists $file] {
		# move file and associated RCS file to the directory
		if {[wtaf_shell_exec "mv -f $file $rcsmv_targetdir" result]} {
			rcsmv $file [file join $rcsmv_targetdir [file tail $file]]
		}
	} else {
		wtaf_err_err $file "No such file or directory"
	}
}



