# This is a -*- Tcl -*- library script and belongs to the
# `Jumble Library for Tcl and Friends'.
#
# error.tcl --- error handling module
#
# Copyright (C) 1996, 1997, 1998 Stefan Hornburg
#
# Author: Stefan Hornburg <racke@gundel.han.de>
# Maintainer: Stefan Hornburg <racke@gundel.han.de>
# Version: see end of file
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
# 
# This file is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this file; see the file COPYING.  If not, write to the Free
# Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

namespace eval Juf {
	namespace eval Error {
		# warning and error messages
		variable _list ""
	}
}

# -----------------------------------------------------------------
# FUNCTION: fault TYPE [ARG ...]
#
# Produces Tcl error.
#
# TYPE describes the error cause and the additional arguments:
#
# wrongargs:	wrong number of arguments (string like "proc arg1")
# badoption:   invalid option (bad option, list with all options)
# -----------------------------------------------------------------

proc Juf::Error::fault {type args} {
	switch -exact -- $type {
		badoption {
			set msg "bad option \"[lindex $args 0]\": must be "
			set v_len [llength [lindex $args 1]]
			set v_list [lsort [lindex $args 1]]
			if {$v_len == 1} {
				append msg $v_list
			} else {
				append msg [join [lrange $v_list 0 [expr $v_len - 2]] ", "]
				append msg ", or " [lindex $v_list end]
			}
		}
		wrongargs {
			set msg "wrong # args: should be \"[lindex $args 0]\""
		}
		default {
			Juf::Error::fatal $type "unknown error type"
		}
	}

	::error $msg
}

# ----------------------------------------------------------------------
# FUNCTION: fatal ARG ...
#
# Concatenates arguments with `:  as separator, prints the resulting
# string on `stderr' and terminates script.
# ----------------------------------------------------------------------

proc Juf::Error::fatal {args} {
	eval Juf::Error::report stderr [lrange $args 0 end]
	exit 1
}

# ----------------------------------------------------------------------
# FUNCTION: error ARG ...
#
# Concatenates arguments with `:  as separator and prints the resulting
# string on `stderr'.
# ----------------------------------------------------------------------

proc Juf::Error::error {args} {
	variable _list
	lappend _list [list error [lrange $args 0 end]]
	eval Juf::Error::report stderr [lrange $args 0 end]
}

# ----------------------------------------------------------------------
# FUNCTION: warning ARG ...
#
# Concatenates string `warning and arguments with `:  as separator
# and prints the resulting string on `stderr'.
# ----------------------------------------------------------------------

proc Juf::Error::warning {args} {
	variable _list
	lappend _list [list warning [lrange $args 0 end]]
	eval Juf::Error::report stderr "warning" [lrange $args 0 end]
}

# ----------------------------------------------------------------
# FUNCTION: summary [TOKEN]
#
# Prints summary of errors/warnings to `stderr' if errors/warnings
# are registered. TOKEN is assumed as source of the problems.
# ----------------------------------------------------------------

proc Juf::Error::summary {{token ""}} {
	variable _list

	if [llength $_list] {
		set warnings 0
		set errors 0
		foreach msg $_list {
			switch -exact -- [lindex $msg 0] warning {
				incr warnings
			} error {
				incr errors
			}
		}
		eval Juf::Error::report stderr $token {
			"$errors error(s), $warnings warning(s)"
		}
	}
}

# -----------------------------------------------------------
# FUNCTION: finish
#
# Terminates script with a exit state of 1, if any errors are
# registered, with 0 otherwise.
# -----------------------------------------------------------

proc Juf::Error::finish {} {
	variable _list
	
	if [llength $_list] {
		foreach msg $_list {
			if {[string compare [lindex $msg 0] error] == 0} {
				exit 1
			} 
		}
	}
	exit 0
}

# -------------------------------------------------
# FUNCTION: tclmsg
#
# Extracts error message from `errorCode' variable.
# Kills script if insufficient information found.
# -------------------------------------------------

proc Juf::Error::tclmsg {} {
	global errorCode
	global errorInfo
	
	switch [lindex $errorCode 0] {
		ARITH -
		NONE {fatal [lindex [split $errorInfo "\n"] 0]}
		POSIX {return [lindex $errorCode 2]}
	}

	return "unknown error"
}

# --------------------------------------------------------
# FUNCTION: report CHANNEL ARG ...
#
# Concatenates arguments with `:  as separator and prints
# the resulting string on CHANNEL.
# --------------------------------------------------------

proc Juf::Error::report {channel args} {

	# Prepend script only if not running interactively
	if {[info script] != ""} {
		set buf [file rootname [file tail [info script]]]
		if {$args != ""} {
			append buf ": "
		}
	}
	append buf [join $args ": "]

	if {$channel == ""} {
		error $buf
	}
	
	puts $channel $buf
}

package provide Error 0.0.7
