# This is a -*- Tcl -*- library script and belongs to the
# `Jumble Library for Tcl and Friends'.
#
# file.tcl --- Manipulating files and filenames
#
# Copyright (C) 1996, 1997, 1998 Stefan Hornburg
#
# Author: Stefan Hornburg <racke@gundel.han.de>
# Maintainer: Stefan Hornburg <racke@gundel.han.de>
# Version: 0.0.5
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
# 
# This file is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this file; see the file COPYING.  If not, write to the Free
# Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package require Jufbase
package require String
package require Sequence
namespace eval Juf {namespace eval File {}}

# ------------------------------------------------------------------
# FUNCTION: expand NAME [DIR]
#
# Makes absolute filename from NAME and performs tilde substitution.
# If NAME is a relative filename, it is located in the current
# working directory or in DIR if given. This function does not
# (yet) remove ".." or "." in filenames.
# ------------------------------------------------------------------

proc Juf::File::expand {name {dir ""}} {
	#
	# absolutify
	#
	if [string compare $dir ""] {
		set fabs [file join $dir $name]
	} else {
		set fabs [file join [pwd] $name]
	}
	#
	# split in parts and perform tilde substitution
	# (see Tcl's file manpage)
	#
	set dlist [file split $fabs]
	set dlist [lreplace $dlist 0 0 [file dirname [lindex $dlist 0]] \
			[file tail [lindex $dlist 0]]]
	#
	# O.K., lbnl merge together
	#
	eval file join $dlist
}

Juf::deprecate juf_file_expand Juf::File::expand

# ------------------------------------------------------------------
# FUNCTION: strip NAME1 NAME2
#
# Returns NAME1 stripped of the common part of both NAME1 and NAME2.
# ------------------------------------------------------------------

proc Juf::File::strip {name1 name2} {
	set dlist1 [file split $name1]
	set dlist2 [file split $name2]

	for {set i 0} "\$i < [llength $dlist1]" {incr i} {
		#
		# Grab leading file name components and compare them
		#
		set dir1 [Juf::Sequence::shift dlist1]
		set dir2 [Juf::Sequence::shift dlist2]
		if {[string compare $dir1 $dir2]} {
			#
			# Different directory name => assemble result
			#
			return [eval file join {$dir1} $dlist1]
		}
	}
	if [llength $dlist1] {
		eval file join $dlist1
	}
}

Juf::deprecate juf_file_strip Juf::File::strip

# ------------------------------------------------
# FUNCTION: class FILE
#
# Uses `file' command to determine class for FILE.
# ------------------------------------------------

proc Juf::File::class {file} {
	set type [file type $file]
	#
	# only regular files should be examined further
	#
	if [string compare $type file] {return $type}
	#
	# run Unix `file' command
	#
	set msg [exec file $file]
	#
	# extract class
	#
	set list [Juf::String::split -- $msg : 2]
	set list [Juf::String::split -- [lindex $list 1] , 2]

	switch -- [string trim [lindex $list 0]] {
		"exported SGML document text" {return sgml}
		"gzip compressed data" {return compressed}
	}
}

Juf::deprecate juf_file_class Juf::File::class

# ------------------------------------------------------------
# FUNCTION: trustopen FILE MODE PERMS
#
# Opens FILE and returns Tcl channel identifier if successful.
# Terminates script otherwise.
# ------------------------------------------------------------

proc Juf::File::trustopen {args} {
	if [catch "open $args" fd] {
		Juf::Error::fatal [lindex $args 0] [Juf::Error::tclmsg]
	}
	set fd
}
	
# -----------------------------------------------------------------
# FUNCTION: slurp FILE NAME
#
# Reads whole FILE into NAME. Returns 1 if successful, 0 otherwise.
# -----------------------------------------------------------------

proc Juf::File::slurp {file name} {
	upvar $name v_buf
	#
	# open file
	#
	if {[catch "open $file" v_chl] != 0} {
		Juf::Error::error $file [Juf::Error::tclmsg]
		return 0
	}
	#
	# read contents
	#
	if {[catch "read $v_chl" v_buf] != 0} {
		Juf::Error::error $file [Juf::Error::tclmsg]
		return 0
	}
	close $v_chl
	return 1
}

Juf::deprecate juf_file_slurp Juf::File::slurp

# -------------------------------------------------------------
# FUNCTION: iscwd NAME
#
# Returns if NAME corresponds to the current working directory,
# 0 otherwise.
# -------------------------------------------------------------

proc Juf::File::iscwd {name} {
	if {[file isdir $name]} {
		set cwd [pwd]	;# remember current working directory
		cd $name		;# move to proposed directory
		if {[pwd] == $cwd} {
			return 1
		} else {
			cd $cwd		;# restore working directory
			return 0
		}
	} else {
		return 0
	}
}

Juf::deprecate juf_file_iscwd Juf::File::iscwd

package provide juf::file 0.0.4
package provide File 0.0.5
