/*
-------------------------------------------------------  
gdbm.c - Accessing GDBM databases with Tcl commands
-------------------------------------------------------

Copyright (C) 1997 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.1

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strdup */
#include <ctype.h> /* tolower */
#include <errno.h>

#include <tcl.h>
#include <gdbm.h>

#include "jufgdbm.h"

/* database information data type */
/* ------------------------------ */

typedef struct {
  char *name ;
  int mode ;
  GDBM_FILE db ;
} Juf_DbInfo ;

/* command information data type */
/* ----------------------------- */

typedef struct {
  const char *name;
  Tcl_CmdProc *cmd;
} Juf_CmdInfo;

/* hash table with handles as keys and database information as values */
/* ------------------------------------------------------------------ */

static Tcl_HashTable juf_gdbm_databases;

/*
 * counter for handles
 */

static juf_gdbm_handle ;



/*
 * prototypes for internal functions
 */

static const char *juf_gdbm_error ();
static Juf_DbInfo *juf_gdbm_getdbinfo (char *handle);

static JUF_TCLDECL(juf_gdbm_close);
static JUF_TCLDECL(juf_gdbm_delete);
static JUF_TCLDECL(juf_gdbm_exists);
static JUF_TCLDECL(juf_gdbm_fetch);
static JUF_TCLDECL(juf_gdbm_insert);
static JUF_TCLDECL(juf_gdbm_list);
static JUF_TCLDECL(juf_gdbm_open);
static JUF_TCLDECL(juf_gdbm_store);

static Juf_CmdInfo juf_gdbm_commands[] = {
  { "close", juf_gdbm_close },
  { "delete", juf_gdbm_delete },
  { "exists", juf_gdbm_exists },
  { "fetch", juf_gdbm_fetch },
  { "insert", juf_gdbm_insert },
  { "list", juf_gdbm_list },
  { "open", juf_gdbm_open },
  { "store", juf_gdbm_store },
  { NULL, NULL}
};

/*
juf_gdbm_error

Returns appropriate error message for current value of
`gdbm_errno'. If `gdbm_errno' is GDBM_FILE_OPEN_ERROR,
the error string for `errno' is returned.
*/
static const char *juf_gdbm_error () {
  static char *errmsg = NULL;
  static const char *sysmsg;
  
  switch (gdbm_errno)
	{
	case GDBM_FILE_OPEN_ERROR: {
	  sysmsg = strerror (errno);
	  break;
	}	
	default: {
	  sysmsg = gdbm_strerror (errno);
	}
	}

  if (errmsg != NULL) free (errmsg);
  errmsg = strdup (sysmsg);
  if (errmsg == NULL) return (strerror (ENOMEM));
  *errmsg = tolower (*errmsg);
  return (errmsg);
}
	  
/*
juf_gdbm_getdbinfo DBID

Returns pointer to structure with database information or
NULL if DBID is invalid.
*/
static Juf_DbInfo *juf_gdbm_getdbinfo (char *handle) {
  	Tcl_HashEntry *entry;
	Juf_DbInfo *dbinfo;
	
	/* search hashtable for database information */
	entry = Tcl_FindHashEntry(&juf_gdbm_databases, handle);
	if (!entry) {return (NULL);}

	/* retrieve pointer to database information */
	dbinfo = Tcl_GetHashValue (entry);
	return (dbinfo);
}

/*
juf_gdbm_close DBID

Closes database specified by handle DBID.
*/

static JUF_TCLDECL(juf_gdbm_close)
{
  Tcl_HashEntry *entry;
  Juf_DbInfo *dbinfo;

  /* check argument number */
  JUF_TCLCKARGS(1, "juf_gdbm close dbId");

  /* search hashtable for database information */
  entry = Tcl_FindHashEntry (&juf_gdbm_databases, JUF_TCLARGX(0));
  if (!entry) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

  /* get database information */
  dbinfo = (Juf_DbInfo *) Tcl_GetHashValue (entry);

  /* close db and remove database information */
  gdbm_close (dbinfo->db) ;
  free (dbinfo->name);
  free ((char*) dbinfo);
  Tcl_DeleteHashEntry(entry);
	
  return (TCL_OK);
}

/*
juf_gdbm_insert DBID KEY CONTENT

Inserts KEY/CONTENT pair into database specified by DBID.
If KEY exists already, an error is generated.
*/	

static JUF_TCLDECL(juf_gdbm_insert)
{
	datum key, content;
	int ret ;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(3, "juf_gdbm insert dbId key newValue");
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo(JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* prepare data and try to store it */
	key.dptr = JUF_TCLARGX(1);
	key.dsize = strlen(JUF_TCLARGX(1)) + 1 ;
	content.dptr = JUF_TCLARGX(2);
	content.dsize = strlen(JUF_TCLARGX(2)) + 1 ;

	ret = gdbm_store (dbinfo -> db, key, content, GDBM_INSERT);
	if (ret != 0) {
	  	Tcl_AppendResult (JUF_TCLIP, gdbm_strerror(gdbm_errno), JUF_TCLEND);
		return TCL_ERROR ;
	}

	return TCL_OK;
}

/*
juf_gdbm_store DBID KEY CONTENT

Inserts KEY/CONTENT pair into database specified by DBID.
If KEY exists already, CONTENT replaces old value.
*/	

static JUF_TCLDECL(juf_gdbm_store)
{
	datum key, content;
	int ret ;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(3, "juf_gdbm store dbId key newValue");
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo(JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* prepare data and try to store it */
	key.dptr = JUF_TCLARGX(1);
	key.dsize = strlen(JUF_TCLARGX(1)) + 1 ;
	content.dptr = JUF_TCLARGX(2);
	content.dsize = strlen(JUF_TCLARGX(2)) + 1 ;

	ret = gdbm_store (dbinfo -> db, key, content, GDBM_REPLACE);
	if (ret != 0) {
	  	Tcl_AppendResult (JUF_TCLIP, gdbm_strerror(gdbm_errno), JUF_TCLEND);
		return TCL_ERROR ;
	}

	return TCL_OK;
}

/*
juf_gdbm_fetch DBID KEY

Searches for KEY in database and returns the corresponding value or
an empty string, if KEY not found.
*/	

static JUF_TCLDECL(juf_gdbm_fetch)
{
	datum key, content;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(2, "juf_gdbm fetch dbId key newValue");
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo(JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* prepare data for key and try to fetch it */
	key.dptr = JUF_TCLARGX(1);
	key.dsize = strlen(JUF_TCLARGX(1)) + 1 ;

	content = gdbm_fetch (dbinfo -> db, key) ;
	if (content.dptr) {
		Tcl_AppendResult (JUF_TCLIP, content.dptr, JUF_TCLEND);
	}
	free (content.dptr);
	
	return TCL_OK ;
}

/*
juf_gdbm_exists DBID KEY

Searches for KEY in database and returns 1 if KEY exists,
0 otherwise.
*/	

static JUF_TCLDECL(juf_gdbm_exists)
{
	datum key;
	int exists;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(2, "juf_gdbm exists dbId key")
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo (JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* prepare data for key and try to fetch it */
	key.dptr = JUF_TCLARGX(1);
	key.dsize = strlen(JUF_TCLARGX(1)) + 1 ;

	exists = gdbm_exists (dbinfo -> db, key) ;
	if (exists) {
		Tcl_SetResult (JUF_TCLIP, "1", TCL_STATIC);
	} else {
	  	Tcl_SetResult (JUF_TCLIP, "0", TCL_STATIC);
	}
	
	return TCL_OK ;
}

/*
juf_gdbm_delete DBID KEY

Removes the data associated with KEY from the database specified by DBID.
*/	

static JUF_TCLDECL(juf_gdbm_delete)
{
	datum key;
	int ret;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(2, "juf_gdbm delete dbId key");
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo(JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* prepare key and try to delete it */
	key.dptr = JUF_TCLARGX(1);
	key.dsize = strlen(JUF_TCLARGX(1)) + 1 ;

	ret = gdbm_delete (dbinfo -> db, key);
	if (ret != 0) {
	  	Tcl_AppendResult (JUF_TCLIP, gdbm_strerror(gdbm_errno), JUF_TCLEND);
		return TCL_ERROR ;
	}

	return TCL_OK;
}

/*
juf_gdbm_list DBID

Returns a list of all keys in the database specified by DBID.
*/	

static JUF_TCLDECL(juf_gdbm_list)
{
	datum key, nextkey;
	Tcl_DString result;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	JUF_TCLCKARGS(1, "juf_gdbm list dbId");
	
	/* get database information */
	dbinfo = juf_gdbm_getdbinfo(JUF_TCLARGX(0));
	if (dbinfo == NULL) {JUF_TCLBADDBHANDLE(JUF_TCLARGX(0))}

	/* loop through keys in database */
	Tcl_DStringInit (&result);

	key = gdbm_firstkey (dbinfo->db);
	if (key.dptr == NULL) {
		/* database empty */
		return TCL_OK ;
	}

	do {
	  /* copy key to result */
	  Tcl_DStringAppendElement (&result, key.dptr);
	  /* fetch next key */
	  nextkey = gdbm_nextkey (dbinfo->db, key);
	  /* free pointer to old key */
	  free (key.dptr);
	  key = nextkey;
	} while (nextkey.dptr != NULL);

	Tcl_DStringResult(interp, &result);
	return TCL_OK ;
}

/*
juf_gdbm_open FILE MODE

Opens FILE as database and returns handle.
MODE is one of r,rw,rwc or rwn.
*/

static JUF_TCLDECL(juf_gdbm_open)
	{
	int mode, new;
	char buf [100];
	Tcl_HashEntry *entry ;
	Juf_DbInfo *dbinfo;
	
	/* check argument number */
	if (JUF_TCLARGC < 2 || JUF_TCLARGC > 3)
	  JUF_TCLWRARGS("juf_gdbm open fileName ?access?");

	mode = GDBM_READER ;

	/* check mode string if given */
	if (JUF_TCLARGC == 3) {
	  if (strcmp(JUF_TCLARGX(1),"r")==0)
		mode = GDBM_READER ;
	  else if (strcmp(JUF_TCLARGX(1),"rw")==0)
		mode = GDBM_WRITER ;
	  else if (strcmp(JUF_TCLARGX(1),"rwc")==0)
		mode = GDBM_WRCREAT ;
	  else if (strcmp(JUF_TCLARGX(1),"rwn")==0)
		mode = GDBM_NEWDB ;
	  else {
		JUF_TCLBADMODE(JUF_TCLARGX(1));
	  }
	}	

	/* open database and register it */
	dbinfo = (Juf_DbInfo *) malloc (sizeof (Juf_DbInfo)) ;
	if (dbinfo == NULL) {
		perror (strerror (ENOMEM));
		exit (1);
	}
	
	/*
	 * create new name and malloc space for it
	 * malloc extra space for name
	 */

	sprintf(buf, "gdbm%i", juf_gdbm_handle);
	juf_gdbm_handle++ ;

	dbinfo -> name = strdup (buf);
	if (dbinfo -> name == NULL) {
	  perror (strerror (ENOMEM));
	  exit (1);
	}

	dbinfo -> mode = mode ;
	dbinfo -> db = gdbm_open (JUF_TCLARGX(0), 0 ,mode, 0664, NULL);

	if (!dbinfo->db) {
		/*
		 * error occurred
		 * free previously allocated memory
		 * db_errno indicates error
		 */
		free (dbinfo->name);
		free ((char*) dbinfo);

		Tcl_AppendResult (JUF_TCLIP, "couldn't open \"", JUF_TCLARGX(0),
						  "\": ", juf_gdbm_error (), JUF_TCLEND);
		return (TCL_ERROR);
	} else {
		/* 
		 * success
		 * enter db to hashtab
		 */
		entry = Tcl_CreateHashEntry(&juf_gdbm_databases,dbinfo->name,&new);
		Tcl_SetHashValue(entry,(char*)dbinfo);
		Tcl_AppendResult (JUF_TCLIP, dbinfo->name, JUF_TCLEND);
		return TCL_OK;
	}
}

/*
juf_gdbm

command dispatcher
*/

JUF_TCLDECL(juf_gdbm)
{
  int i;
  
  /* check argument number */
  if (JUF_TCLARGC < 3) JUF_TCLWRARGS("juf_gdbm option arg ?arg ...?");

  /* select appropriate function */
  for (i = 0 ; juf_gdbm_commands[i].name != NULL; i++) {
	if (strcmp (juf_gdbm_commands[i].name, JUF_TCLARGX (0)) == 0)
	  {
		return (juf_gdbm_commands [i].cmd
				(JUF_TCLCD, JUF_TCLIP, JUF_TCLARGC - 1, &JUF_TCLARGX(0)));
	  }
  }	
	
  /* command not found */	
  Tcl_AppendResult (JUF_TCLIP, "bad option \"", JUF_TCLARGX(0),
					"\": should be close, delete, exists, fetch, insert, list, open, or store", JUF_TCLEND);
  return(TCL_ERROR);
}

/*
Jufgdbm_Init INTERP

Initializes library.
*/

int Jufgdbm_Init (Tcl_Interp *JUF_TCLIP)
	{	
	Tcl_CreateCommand(JUF_TCLIP, "juf_gdbm", juf_gdbm,
			  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	/* init hash table */
	/* --------------- */

	Tcl_InitHashTable (&juf_gdbm_databases, TCL_STRING_KEYS);

	/*
	 * set handle to zero
	 */

	juf_gdbm_handle = 0 ;

	/* provide JUFGDBM as package */
	return (Tcl_PkgProvide (JUF_TCLIP, JUFGDBM_PKGNAME, JUFGDBM_VERSION));
}
