/*
-------------------------------------------------------  
msql.c - Accessing mSQL databases with Tcl commands
-------------------------------------------------------

Copyright (C) 1998 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.9

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strdup */
#include <ctype.h> /* tolower */
#include <errno.h>
#include <time.h> /* needed for msql.h */

#include <tcl.h>
#include <msql.h>

#include "juf.h"

/* data about a database connection */
typedef struct jufmsql_conndata {
	int handle;        /* Msql internal handle for connection */
	int conn;      /* the connection to the mSQL backend */
} jufmsql_conndata;
/* corresponding Tcl object type */
Tcl_ObjType *jufmsql_type;

/* maps connection handles to corresponding data */
static Tcl_HashTable jufmsql_databases;
/* counter for connection handles */
static jufmsql_handle = 0;

/* Tcl object type for results retrieved from POSTGRES */
Tcl_ObjType *jufmsql_result_type;
/* maps result handles to corresponding data */
static Tcl_HashTable jufmsql_results;

/* prototypes */
int jufmsql_lookup (Tcl_Interp *interp, Tcl_Obj *objptr);
void jufmsql_tostring (Tcl_Obj *obj);
void jufmsql_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj);

int jufmsql_result_lookup (Tcl_Interp *interp, Tcl_Obj *objptr);
void jufmsql_result_tostring (Tcl_Obj *obj);
void jufmsql_result_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj);

int jufmsql_connect (ClientData data, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[]);
int jufmsql_exec (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
int jufmsql_tuples (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
int jufmsql_value (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
void jufmsql_cleanse (ClientData data);
					 
/*
 * --------------------------------------------------
 * FUNCTION: jufmsql_lookup OBJ
 *
 * Updates internal representation for Tcl objects of
 * the PGConn type.
 * --------------------------------------------------
 */

int jufmsql_lookup (Tcl_Interp *interp, Tcl_Obj *objptr)
{
	Tcl_HashEntry *entry;
	Tcl_Obj *buf;
  
	entry = Tcl_FindHashEntry (&jufmsql_databases,
							   Tcl_GetStringFromObj (objptr, NULL));
	if (entry == NULL)
		{
		if (interp != NULL)
			{
			/* error message mandatory */
			buf = Tcl_NewStringObj ("", 0);
			Tcl_AppendStringsToObj (buf, "no such connection \"", 
									Tcl_GetStringFromObj (objptr, NULL),
									"\"", NULL);
			Tcl_SetObjResult (interp, buf);
			}
		return (TCL_ERROR);
		}
	else
		{
		/* free old internal representation if valid and handler is defined */
		if (objptr -> typePtr != NULL
			&& objptr -> typePtr -> freeIntRepProc != NULL)
			{
			objptr -> typePtr -> freeIntRepProc (objptr);
			}
		objptr -> internalRep.otherValuePtr = Tcl_GetHashValue (entry);
		objptr -> typePtr = jufmsql_type;
		return (TCL_OK);
		}
}

/*
 * ----------------------------------------------------
 * FUNCTION: jufmsql_tostring TCLOBJ
 *
 * Updates string representation for Tcl objects of the
 * PGConn type.
 * ----------------------------------------------------
 */

void jufmsql_tostring (Tcl_Obj *obj) {
	char buf[100];
	jufmsql_conndata *conndata =
		(jufmsql_conndata *) obj -> internalRep.otherValuePtr;
  
	sprintf (buf, "msql%i", conndata -> handle);
	obj -> bytes = Tcl_Alloc (((unsigned int) (strlen (buf) + 1)));
	obj -> length = strlen (buf);
	strcpy (obj -> bytes, buf);
}

/* jufmsql_dup FROMOBJ TOOBJ
 *
 * Duplicates FROMOBJ and stores the result into TOOBJ.
 */

void jufmsql_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj) {
	toobj -> internalRep.otherValuePtr = fromobj -> internalRep.otherValuePtr;
}

/*
 * -------------------------------------------
 * FUNCTION: jufmsql_result_lookup INTERP OBJ
 *
 * Searches for connection represented by OBJ.
 * -------------------------------------------
 */

int jufmsql_result_lookup (Tcl_Interp *interp, Tcl_Obj *objptr)
{
	Tcl_HashEntry *entry;
	Tcl_Obj *buf;
  
	entry = Tcl_FindHashEntry (&jufmsql_results,
							   Tcl_GetStringFromObj (objptr, NULL));
	if (entry == NULL)
		{
		buf = Tcl_NewStringObj ("", 0);
		Tcl_AppendStringsToObj (buf, "invalid result handle \"", 
								Tcl_GetStringFromObj (objptr, NULL),
								"\"", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	else
		{
  		/* free old internal representation if valid and handler is defined */
		if (objptr -> typePtr != NULL &&
			objptr -> typePtr -> freeIntRepProc != NULL)
			{
			objptr -> typePtr -> freeIntRepProc (objptr);
			}
		objptr -> internalRep.otherValuePtr = Tcl_GetHashValue (entry);
		objptr -> typePtr = jufmsql_result_type;
		return (TCL_OK);
		}
}

/*
 * --------------------------------------
 * FUNCTION: jufmsql_result_tostring OBJ
 *
 * Creates string representation for OBJ.
 * --------------------------------------
 */

void jufmsql_result_tostring (Tcl_Obj *obj)
{
	obj -> length = sizeof (void *) * 2; 
	obj -> bytes = Tcl_Alloc ((unsigned int) (obj -> length + 1));
	sprintf (obj -> bytes, "%p", obj -> internalRep.otherValuePtr);
}

/*
 * jufmsql_result_dup FROMOBJ TOOBJ
 *
 * Duplicates FROMOBJ and stores the result into TOOBJ.
 */

void jufmsql_result_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj) {
  toobj -> internalRep.otherValuePtr = fromobj -> internalRep.otherValuePtr;
}

/*
 * ---------------------------------------------------------
 * COMMAND: jufmsql_connect DBNAME [HOST]
 *
 * Creates a new connection to a mSQL backend server.
 * ---------------------------------------------------------
 */

int jufmsql_connect (ClientData data, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[])
{
	Tcl_Obj *obuf = Tcl_NewStringObj ("", 0);
	jufmsql_conndata *conndata;
	char buf[100];
	Tcl_HashEntry *entry;
	int flag; int retval; int conn;
	char *dbname = NULL;
	char *host = NULL;
	Tcl_Obj *msql_objv[3];
	int msql_objc = 0;
	Tcl_Obj *msql_new;
  
	/* check argument number */
	if (objc > 3)
		{
		Tcl_AppendStringsToObj (obuf, "wrong # args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" dbname ?host?", NULL);
		Tcl_SetObjResult (interp, obuf);
		return(TCL_ERROR);
		}

	/* setup arguments for postgres */
	switch (objc)
		{
		case 3:
			host = Tcl_GetStringFromObj (objv[2], NULL);
			msql_objv[1] = Tcl_DuplicateObj (objv[2]);
			msql_objc++;
		case 2:
			dbname = Tcl_GetStringFromObj (objv[1], NULL);
			msql_objv[0] = Tcl_DuplicateObj (objv[1]);
			msql_objc++;
		}
 
	/* try to establish connection */
	conn = msqlConnect (host);
	if (conn == -1)
		{
		/* failure */
		Tcl_AppendStringsToObj (obuf, msqlErrMsg, NULL);
		Tcl_SetObjResult (interp, obuf);
		return (TCL_ERROR);
		}

	/* select database */
	retval = msqlSelectDB (conn, dbname);
	if (retval == -1)
		{
		/* failure */
		Tcl_AppendStringsToObj (obuf, msqlErrMsg, NULL);
		Tcl_SetObjResult (interp, obuf);
		return (TCL_ERROR);
		}
	
	/* success => record connection data */
	conndata = malloc (sizeof (jufmsql_conndata));
	conndata -> conn = conn;
	conndata -> handle = jufmsql_handle;
	sprintf (buf, "msql%i", jufmsql_handle++);
	entry = Tcl_CreateHashEntry (&jufmsql_databases, buf, &flag);
	Tcl_SetHashValue (entry, conndata);

	/* create new object */
	msql_new = Tcl_NewObj ();
	msql_new -> internalRep.otherValuePtr = conndata;
	msql_new -> typePtr = jufmsql_type;
	msql_new -> bytes = NULL;
	Tcl_IncrRefCount (msql_new);
	
	/* set result to database handler */
	Tcl_SetObjResult (interp, msql_new);
	return (TCL_OK);
}

/*
 * ------------------------------------------
 * COMMAND: jufmsql_exec CONN QUERY
 *
 * Submit a query to postgres.
 * ------------------------------------------
 */

int jufmsql_exec (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	jufmsql_conndata *conndata;
	m_result *exec_result;
	char resrep [sizeof (void *) * 2 + 1];
	char *query;
	int flag; int result;
  
	/* check argument number */
	if (objc != 3)
		{
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" conn query", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufmsql_type) != TCL_OK)
		{
		/* not a valid connection representation */
		return (TCL_ERROR);
		}
	
	conndata = objv[1] -> internalRep.otherValuePtr;
	query = Tcl_GetStringFromObj (objv[2], NULL);
	result = msqlQuery (conndata -> conn, query);

	if (result == -1)
		{
		/* query failed */
		Tcl_AppendStringsToObj (buf, "Query \"", query, "\" failed\n",
								msqlErrMsg, NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	exec_result = msqlStoreResult ();
	
	/* record result */
	sprintf (resrep, "%p", exec_result);
	Tcl_CreateHashEntry (&jufmsql_results, resrep, &flag);
	buf = Tcl_NewObj ();
	buf -> internalRep.otherValuePtr = exec_result;
	buf -> typePtr = jufmsql_result_type;
	Tcl_SetObjResult (interp, buf);
	return (TCL_OK);
}

/*
 * jufmsql_tuples RESULT
 *
 * Returns number of tuples within RESULT.
 */

int jufmsql_tuples (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[]) {
  Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
  m_result *result;
  
  /* check argument number */
  if (objc != 2) {
	Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " result", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }  

  if (Tcl_ConvertToType (interp, objv[1], jufmsql_result_type) == TCL_ERROR) {
	/* not a valid connection representation */
	return (TCL_ERROR);
  }

  result = objv[1] -> internalRep.otherValuePtr;
  Tcl_SetObjResult (interp, Tcl_NewIntObj (msqlNumRows (result)));
  return (TCL_OK);
}

/*
 * jufmsql_value RESULT TUPLE FIELD
 *
 * Returns value of FIELD within TUPLE within RESULT.
 */

int jufmsql_value (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[]) {
  Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
  m_result *result; m_field *field; m_row dbdata;
  int tuple; int tuples; int fields; int i;
  char *name;
  
  /* check argument number */
  if (objc != 4) {
	Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " result tuple field", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }  

  if (Tcl_ConvertToType (interp, objv[1], jufmsql_result_type) == TCL_ERROR) {
	/* not a valid connection representation */
	return (TCL_ERROR);
  }
  
  result = objv[1] -> internalRep.otherValuePtr;

  /* tuple O.K. ? */
  if (Tcl_GetIntFromObj (interp, objv[2], &tuple) == TCL_ERROR) {
	return (TCL_ERROR);
  }
  tuples = msqlNumRows (result);
  if (tuple < 0 || tuple >= tuples) {
	Tcl_AppendStringsToObj (buf, "tuple out of range", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }

  /* select and retrieve tuple */
  msqlDataSeek (result, tuple);
  dbdata = msqlFetchRow (result);
	  
  /* search field */
  fields = msqlNumFields (result);
  name = Tcl_GetStringFromObj (objv[3], NULL);
  for (i = 0; i < fields; i++)
	  {
	  /* select field and retrieve data */
	  msqlFieldSeek (result, i);
	  field = msqlFetchField (result);

	  if (field == NULL)
		  {
		  /* something went awry */
		  Tcl_AppendStringsToObj (buf, "couldn't fetch field", NULL);
		  Tcl_SetObjResult (interp, buf);
		  return (TCL_ERROR);
		  }

	  if (strcmp (name, field -> name) == 0)
		  {		  
		  Tcl_SetObjResult (interp, Tcl_NewStringObj (dbdata[i],
													  field -> length));
		  return (TCL_OK);
		  }
	  }

  Tcl_AppendStringsToObj (buf, "field \"", name, "\" not found", NULL);
  Tcl_SetObjResult (interp, buf);
  return (TCL_ERROR);
}

/*
 * ------------------------------------------
 * FUNCTION: jufmsql_cleanse DATA
 *
 * Closes unfinished connections and results.
 * ------------------------------------------
 */

void jufmsql_cleanse (ClientData data) {
  Tcl_HashSearch hsrch;
  Tcl_HashEntry *entry;
  jufmsql_conndata *conndata;
  m_result *result;
  
  entry = Tcl_FirstHashEntry (&jufmsql_databases, &hsrch);
  while (entry != NULL) {
	conndata = (jufmsql_conndata *) Tcl_GetHashValue (entry);
	msqlClose (conndata -> conn);
	free (conndata);
	entry = Tcl_NextHashEntry (&hsrch);
  }

  entry = Tcl_FirstHashEntry (&jufmsql_results, &hsrch);
  while (entry != NULL) {
	result = (m_result *) Tcl_GetHashValue (entry);
	msqlFreeResult (result);
	entry = Tcl_NextHashEntry (&hsrch);
  }
}
 
/*
Jufmsql_Init INTERP

Initializes library.
*/

int Jufmsql_Init (Tcl_Interp *interp) {

  if (jufmsql_type == NULL) {
	/* create new Tcl object type */
	jufmsql_type = (Tcl_ObjType *) Tcl_Alloc (sizeof (Tcl_ObjType));	
	jufmsql_type -> name = "PGconn";
	jufmsql_type -> freeIntRepProc = NULL;
	jufmsql_type -> dupIntRepProc = jufmsql_dup;
	jufmsql_type -> updateStringProc = jufmsql_tostring;
	jufmsql_type -> setFromAnyProc = jufmsql_lookup;
	Tcl_RegisterObjType (jufmsql_type);
  }
  if (jufmsql_result_type == NULL) {
	/* create new Tcl object type */
	jufmsql_result_type = (Tcl_ObjType *) Tcl_Alloc (sizeof (Tcl_ObjType));	
	jufmsql_result_type -> name = "PGresult";
	jufmsql_result_type -> freeIntRepProc = NULL;
	jufmsql_result_type -> dupIntRepProc = jufmsql_result_dup;
	jufmsql_result_type -> updateStringProc = jufmsql_result_tostring;
	jufmsql_result_type -> setFromAnyProc = jufmsql_result_lookup;
	Tcl_RegisterObjType (jufmsql_result_type);
  }
  
  /* create Tcl commands */
  Tcl_CreateObjCommand (interp, "Juf::Msql::connect", jufmsql_connect,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Msql::exec", jufmsql_exec,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Msql::tuples", jufmsql_tuples,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Msql::value", jufmsql_value,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  /* init hash tables */
  Tcl_InitHashTable (&jufmsql_databases, TCL_STRING_KEYS);
  Tcl_InitHashTable (&jufmsql_results, TCL_STRING_KEYS);
					 
  /* install exit handler */
  Tcl_CreateExitHandler (jufmsql_cleanse, (ClientData) NULL);

  /* provide JUFMSQL as package */
  return (Tcl_PkgProvide (interp, JUFMSQL_PKGNAME, JUFMSQL_VERSION));
}

