/*
-------------------------------------------------------  
pq.c - Accessing POSTGRES databases with Tcl commands
-------------------------------------------------------

Copyright (C) 1997, 1998 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.9

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strdup */
#include <ctype.h> /* tolower */
#include <errno.h>

#include <tcl.h>
#include <libpq-fe.h>

#include "juf.h"

/* data about a database connection */
typedef struct jufpq_conndata {
	int handle;        /* Pq internal handle for connection */
	PGconn *conn;      /* the connection to the postgres backend */
} jufpq_conndata;
/* corresponding Tcl object type */
Tcl_ObjType *jufpq_type;

/* maps connection handles to corresponding data */
static Tcl_HashTable jufpq_databases;
/* counter for connection handles */
static jufpq_handle = 0;

/* Tcl object type for results retrieved from POSTGRES */
Tcl_ObjType *jufpq_result_type;
/* maps result handles to corresponding data */
static Tcl_HashTable jufpq_results;

/* prototypes */
int jufpq_lookup (Tcl_Interp *interp, Tcl_Obj *objptr);
void jufpq_tostring (Tcl_Obj *obj);
void jufpq_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj);

int jufpq_result_lookup (Tcl_Interp *interp, Tcl_Obj *objptr);
void jufpq_result_tostring (Tcl_Obj *obj);
void jufpq_result_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj);

int jufpq_connect (ClientData data, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[]);
int jufpq_exec (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
int jufpq_tuples (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
int jufpq_value (ClientData data, Tcl_Interp *interp,  int objc,
				   Tcl_Obj *CONST objv[]);
void jufpq_cleanse (ClientData data);

/* Auxiliary functions */
/* =================== */

/*
 * -------------------------------------------------------------
 * FUNCTION: jufpq_massage_errmsg INTERP STRING
 *
 * Strips trailing newlines and error classifiers from an
 * error message and returns the result (dynamically allocated).
 *
 * Example:
 * "FATAL 1:  Database arte does not exist in pg_database
 * "
 * => "Database arte does not exist in pg_database"
 * -------------------------------------------------------------
 */

char *jufpq_massage_errmsg (Tcl_Interp *interp, const char *string)
{
	char *newstring = strdup (string);
	int matchres;
	Tcl_RegExp regex;
	
	/* compile regular expression */
	if ((regex = Tcl_RegExpCompile (interp, "^FATAL [0-9]*: *")) == NULL)
		{
		/* bug */
		fprintf (stderr, interp -> result);
		exit (1);
		}
		
}

/*
 * --------------------------------------------------
 * FUNCTION: jufpq_lookup OBJ
 *
 * Updates internal representation for Tcl objects of
 * the PGConn type.
 * --------------------------------------------------
 */

int jufpq_lookup (Tcl_Interp *interp, Tcl_Obj *objptr)
{
	Tcl_HashEntry *entry;
	Tcl_Obj *buf;
  
	entry = Tcl_FindHashEntry (&jufpq_databases,
							   Tcl_GetStringFromObj (objptr, NULL));
	if (entry == NULL)
		{
		if (interp != NULL)
			{
			/* error message mandatory */
			buf = Tcl_NewStringObj ("", 0);
			Tcl_AppendStringsToObj (buf, "no such connection \"", 
									Tcl_GetStringFromObj (objptr, NULL),
									"\"", NULL);
			Tcl_SetObjResult (interp, buf);
			}
		return (TCL_ERROR);
		}
	else
		{
		/* free old internal representation if valid and handler is defined */
		if (objptr -> typePtr != NULL
			&& objptr -> typePtr -> freeIntRepProc != NULL)
			{
			objptr -> typePtr -> freeIntRepProc (objptr);
			}
		objptr -> internalRep.otherValuePtr = Tcl_GetHashValue (entry);
		objptr -> typePtr = jufpq_type;
		return (TCL_OK);
		}
}

/*
 * ----------------------------------------------------
 * FUNCTION: jufpq_tostring TCLOBJ
 *
 * Updates string representation for Tcl objects of the
 * PGConn type.
 * ----------------------------------------------------
 */

void jufpq_tostring (Tcl_Obj *obj) {
	char buf[100];
	jufpq_conndata *conndata =
		(jufpq_conndata *) obj -> internalRep.otherValuePtr;
  
	sprintf (buf, "pq%i", conndata -> handle);
	obj -> bytes = Tcl_Alloc (((unsigned int) (strlen (buf) + 1)));
	obj -> length = strlen (buf);
	strcpy (obj -> bytes, buf);
}

/* jufpq_dup FROMOBJ TOOBJ
 *
 * Duplicates FROMOBJ and stores the result into TOOBJ.
 */

void jufpq_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj) {
	toobj -> internalRep.otherValuePtr = fromobj -> internalRep.otherValuePtr;
}

/*
 * -------------------------------------------
 * FUNCTION: jufpq_result_lookup INTERP OBJ
 *
 * Searches for connection represented by OBJ.
 * -------------------------------------------
 */

int jufpq_result_lookup (Tcl_Interp *interp, Tcl_Obj *objptr)
{
	Tcl_HashEntry *entry;
	Tcl_Obj *buf;
  
	entry = Tcl_FindHashEntry (&jufpq_results,
							   Tcl_GetStringFromObj (objptr, NULL));
	if (entry == NULL)
		{
		buf = Tcl_NewStringObj ("", 0);
		Tcl_AppendStringsToObj (buf, "invalid result handle \"", 
								Tcl_GetStringFromObj (objptr, NULL),
								"\"", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	else
		{
  		/* free old internal representation if valid and handler is defined */
		if (objptr -> typePtr != NULL &&
			objptr -> typePtr -> freeIntRepProc != NULL)
			{
			objptr -> typePtr -> freeIntRepProc (objptr);
			}
		objptr -> internalRep.otherValuePtr = Tcl_GetHashValue (entry);
		objptr -> typePtr = jufpq_result_type;
		return (TCL_OK);
		}
}

/*
 * --------------------------------------
 * FUNCTION: jufpq_result_tostring OBJ
 *
 * Creates string representation for OBJ.
 * --------------------------------------
 */

void jufpq_result_tostring (Tcl_Obj *obj)
{
	obj -> length = sizeof (void *) * 2; 
	obj -> bytes = Tcl_Alloc ((unsigned int) (obj -> length + 1));
	sprintf (obj -> bytes, "%p", obj -> internalRep.otherValuePtr);
}

/*
 * ----------------------------------------------------
 * FUNCTION: jufpq_result_dup FROMOBJ TOOBJ
 *
 * Duplicates FROMOBJ and stores the result into TOOBJ.
 * ----------------------------------------------------
 */

void jufpq_result_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj) {
  toobj -> internalRep.otherValuePtr = fromobj -> internalRep.otherValuePtr;
}

/*
 * ---------------------------------------------------------
 * COMMAND: jufpq_connect [DBNAME HOST PORT OPTIONS LOGFILE]
 *
 * Creates a new connection to a postgres backend server.
 * ---------------------------------------------------------
 */

int jufpq_connect (ClientData data, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[])
{
	Tcl_Obj *obuf = Tcl_NewStringObj ("", 0);
	PGconn *conn;
	jufpq_conndata *conndata;
	char buf[100];
	Tcl_HashEntry *entry;
	int flag;
	char *dbname = NULL;
	char *host = NULL;
	char *port = NULL;
	char *options = NULL;
	char *logfile = NULL;
	Tcl_Obj *pq_objv[3];
	int pq_objc = 0;
	Tcl_Obj *pq_new;
  
	/* check argument number */
	if (objc > 6)
		{
		Tcl_AppendStringsToObj (obuf, "wrong # args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" ?dbname host port options logfile?", NULL);
		Tcl_SetObjResult (interp, obuf);
		return(TCL_ERROR);
		}

	/* setup arguments for postgres */
	switch (objc)
		{
		case 6: logfile = Tcl_GetStringFromObj (objv[5], NULL);
		case 5: options = Tcl_GetStringFromObj (objv[4], NULL);
		case 4:
			port = Tcl_GetStringFromObj (objv[3], NULL);
			pq_objv[2] = Tcl_DuplicateObj (objv[3]);
			pq_objc++;
		case 3:
			host = Tcl_GetStringFromObj (objv[2], NULL);
			pq_objv[1] = Tcl_DuplicateObj (objv[2]);
			pq_objc++;
		case 2:
			dbname = Tcl_GetStringFromObj (objv[1], NULL);
			pq_objv[0] = Tcl_DuplicateObj (objv[1]);
			pq_objc++;
		}
 
	/* try to establish connection */
	conn = PQsetdb (host, port, options, logfile, dbname);
	if (PQstatus (conn) == CONNECTION_OK)
		{
		/* success => record connection data */
		conndata = malloc (sizeof (jufpq_conndata));
		conndata -> conn = conn;
		conndata -> handle = jufpq_handle;
		sprintf (buf, "pq%i", jufpq_handle++);
		entry = Tcl_CreateHashEntry (&jufpq_databases, buf, &flag);
		Tcl_SetHashValue (entry, conndata);

		/* create new object */
		pq_new = Tcl_NewObj ();
		pq_new -> internalRep.otherValuePtr = conndata;
		pq_new -> typePtr = jufpq_type;
		pq_new -> bytes = NULL;
		Tcl_IncrRefCount (pq_new);

		/* set result to database handler */
		Tcl_SetObjResult (interp, pq_new);
		return (TCL_OK);
		}
	else
		{
		/* failure */
		Tcl_AppendStringsToObj (obuf, PQerrorMessage (conn), NULL);
		Tcl_SetObjResult (interp, obuf);
		return (TCL_ERROR);
		}
}

/*
 * ------------------------------------------
 * COMMAND: jufpq_exec CONN QUERY
 *
 * Submit a query to postgres.
 * ------------------------------------------
 */

int jufpq_exec (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	jufpq_conndata *conndata;
	PGresult *result;
	char resrep [sizeof (void *) * 2 + 1];
	char *query;
	int flag;
  
	/* check argument number */
	if (objc != 3)
		{
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" conn query", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufpq_type) != TCL_OK)
		{
		/* not a valid connection representation */
		return (TCL_ERROR);
		}
	
	conndata = objv[1] -> internalRep.otherValuePtr;
	query = Tcl_GetStringFromObj (objv[2], NULL);
	result = PQexec (conndata -> conn, query);

	if (result == NULL)
		{
		/* query failed */
		Tcl_AppendStringsToObj (buf, "Query \"", query, "\" failed\n",
								PQerrorMessage (conndata -> conn), NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	
	switch (PQresultStatus (result))
		{
		case PGRES_COMMAND_OK:
			PQclear (result);
			Tcl_ResetResult (interp);
			return (TCL_OK);
		case PGRES_TUPLES_OK:
			/* record result */
			sprintf (resrep, "%p", result);
			Tcl_CreateHashEntry (&jufpq_results, resrep, &flag);
			buf = Tcl_NewObj ();
			buf -> internalRep.otherValuePtr = result;
			buf -> typePtr = jufpq_result_type;
			Tcl_SetObjResult (interp, buf);
			return (TCL_OK);
		case PGRES_EMPTY_QUERY:
			Tcl_AppendStringsToObj (buf, "empty query", NULL);
			Tcl_SetObjResult (interp, buf);
			PQclear (result);
			return (TCL_ERROR);
		case PGRES_NONFATAL_ERROR:
		case PGRES_FATAL_ERROR:
			Tcl_AppendStringsToObj (buf, PQerrorMessage (conndata -> conn),
									"Query \"", query, "\" failed", NULL);
			Tcl_SetObjResult (interp, buf);
			PQclear (result);
			return (TCL_ERROR);
		default:
			Tcl_AppendStringsToObj (buf, "query failed",
									PQcmdStatus (result), NULL);
			Tcl_SetObjResult (interp, buf);
			PQclear (result);
			return (TCL_ERROR);
		}
}

/*
 * jufpq_tuples RESULT
 *
 * Returns number of tuples within RESULT.
 */

int jufpq_tuples (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[]) {
  Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
  PGresult *result;
  
  /* check argument number */
  if (objc != 2) {
	Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " result", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }  

  if (Tcl_ConvertToType (interp, objv[1], jufpq_result_type) == TCL_ERROR) {
	/* not a valid connection representation */
	return (TCL_ERROR);
  }

  result = objv[1] -> internalRep.otherValuePtr;
  Tcl_SetObjResult (interp, Tcl_NewIntObj (PQntuples (result)));
  return (TCL_OK);
}

/*
 * jufpq_value RESULT TUPLE FIELD
 *
 * Returns value of FIELD within TUPLE within RESULT.
 */

int jufpq_value (ClientData data, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[]) {
  Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
  PGresult *result; int tuple; int tuples; int fields; int i;
  char *name; char *value;
  
  /* check argument number */
  if (objc != 4) {
	Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " result tuple field", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }  

  if (Tcl_ConvertToType (interp, objv[1], jufpq_result_type) == TCL_ERROR) {
	/* not a valid connection representation */
	return (TCL_ERROR);
  }
  
  result = objv[1] -> internalRep.otherValuePtr;

  /* tuple O.K. ? */
  if (Tcl_GetIntFromObj (interp, objv[2], &tuple) == TCL_ERROR) {
	return (TCL_ERROR);
  }
  tuples = PQntuples (result);
  if (tuple < 0 || tuple >= tuples) {
	Tcl_AppendStringsToObj (buf, "tuple out of range", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }
  
  /* search field */
  fields = PQnfields (result);
  name = Tcl_GetStringFromObj (objv[3], NULL);
  for (i = 0; i < fields; i++) {
	if (strcmp (name, PQfname (result, i)) == 0) {
	  value = PQgetvalue (result, tuple, i);
	  Tcl_SetObjResult (interp, Tcl_NewStringObj (value,
												  (int) strlen (value)));
	  return (TCL_OK);
	}
  }
  Tcl_AppendStringsToObj (buf, "field \"", name, "\" not found", NULL);
  Tcl_SetObjResult (interp, buf);
  return (TCL_ERROR);
}

/*
 * ------------------------------------------
 * FUNCTION: jufpq_cleanse DATA
 *
 * Closes unfinished connections and results.
 * ------------------------------------------
 */

void jufpq_cleanse (ClientData data) {
  Tcl_HashSearch hsrch;
  Tcl_HashEntry *entry;
  jufpq_conndata *conndata;
  PGresult *result;
  
  entry = Tcl_FirstHashEntry (&jufpq_databases, &hsrch);
  while (entry != NULL) {
	conndata = (jufpq_conndata *) Tcl_GetHashValue (entry);
	PQfinish (conndata -> conn);
	free (conndata);
	entry = Tcl_NextHashEntry (&hsrch);
  }

  entry = Tcl_FirstHashEntry (&jufpq_results, &hsrch);
  while (entry != NULL) {
	result = (PGresult *) Tcl_GetHashValue (entry);
	PQclear (result);
	entry = Tcl_NextHashEntry (&hsrch);
  }
}
 
/*
Jufpq_Init INTERP

Initializes library.
*/

int Jufpq_Init (Tcl_Interp *interp) {

  if (jufpq_type == NULL) {
	/* create new Tcl object type */
	jufpq_type = (Tcl_ObjType *) Tcl_Alloc (sizeof (Tcl_ObjType));	
	jufpq_type -> name = "PGconn";
	jufpq_type -> freeIntRepProc = NULL;
	jufpq_type -> dupIntRepProc = jufpq_dup;
	jufpq_type -> updateStringProc = jufpq_tostring;
	jufpq_type -> setFromAnyProc = jufpq_lookup;
	Tcl_RegisterObjType (jufpq_type);
  }
  if (jufpq_result_type == NULL) {
	/* create new Tcl object type */
	jufpq_result_type = (Tcl_ObjType *) Tcl_Alloc (sizeof (Tcl_ObjType));	
	jufpq_result_type -> name = "PGresult";
	jufpq_result_type -> freeIntRepProc = NULL;
	jufpq_result_type -> dupIntRepProc = jufpq_result_dup;
	jufpq_result_type -> updateStringProc = jufpq_result_tostring;
	jufpq_result_type -> setFromAnyProc = jufpq_result_lookup;
	Tcl_RegisterObjType (jufpq_result_type);
  }
  
  /* create Tcl commands */
  Tcl_CreateObjCommand (interp, "Juf::Pq::connect", jufpq_connect,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Pq::exec", jufpq_exec,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Pq::tuples", jufpq_tuples,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateObjCommand (interp, "Juf::Pq::value", jufpq_value,
					 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  
  /* init hash tables */
  Tcl_InitHashTable (&jufpq_databases, TCL_STRING_KEYS);
  Tcl_InitHashTable (&jufpq_results, TCL_STRING_KEYS);
					 
  /* install exit handler */
  Tcl_CreateExitHandler (jufpq_cleanse, (ClientData) NULL);

  /* provide JUFPQ as package */
  return (Tcl_PkgProvide (interp, JUFPQ_PKGNAME, JUFPQ_VERSION));
}

