/*
------------------------------------------------
juftime.c - Auxiliary functions related to times
------------------------------------------------

Copyright (C) 1997, 1998 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.9

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <tcl.h>
#include <sys/times.h>
#include <time.h>
#include "juftime.h"

void
TclXOSElapsedTime (clock_t *realTime, clock_t *cpuTime);

/*
 * Time information
 */

jufTimeEntry_t *JufTimerAlloc (void)
{
	jufTimeEntry_t *pTm = (jufTimeEntry_t *)
		Tcl_Alloc (sizeof (jufTimeEntry_t));
	pTm -> real_sum = 0;
	pTm -> cpu_sum = 0;
	return (pTm);
}

void JufTimerStart (jufTimeEntry_t *pTm) 
{
  pTm -> real_sum = 0;
  pTm -> cpu_sum = 0;
  TclXOSElapsedTime (& (pTm -> real_last), & (pTm -> cpu_last));
}

void JufTimerSuspend (jufTimeEntry_t *pTm)
{
  clock_t real, cpu;

  if (pTm -> real_last && pTm -> cpu_last)
	  {
	  TclXOSElapsedTime (&real, &cpu);
	  pTm -> real_sum += (real - pTm -> real_last);
	  pTm -> cpu_sum += (cpu - pTm -> cpu_last);
	  pTm -> real_last = 0;
	  pTm -> cpu_last = 0;
	  }
}

void JufTimerCont (jufTimeEntry_t *pTm)
{
  TclXOSElapsedTime (& (pTm -> real_last), & (pTm -> cpu_last));
}

/*
 * Excerpt from unixos.c (TclX distribution)
 *-----------------------------------------------------------------------------
 * Copyright 1996-1996 Karl Lehenbauer and Mark Diekhans.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies.  Karl Lehenbauer and
 * Mark Diekhans make no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without express or
 * implied warranty.
 *-----------------------------------------------------------------------------
 * $Id: juftime.c,v 1.11 1998/11/20 00:02:54 racke Exp racke $
 *-----------------------------------------------------------------------------
 */

/*-----------------------------------------------------------------------------
 * TclXOSTicksToMS --
 *
 *   Convert clock ticks to milliseconds.
 *
 * Parameters:
 *   o numTicks - Number of ticks.
 * Returns:
 *   Milliseconds.
 *-----------------------------------------------------------------------------
 */
clock_t
TclXOSTicksToMS (numTicks)
    clock_t numTicks;
{
    static clock_t msPerTick = 0;

    /*
     * Some systems (SVR4) implement CLK_TCK as a call to sysconf, so lets only
     * reference it once in the life of this process.
     */
    if (msPerTick == 0)
        msPerTick = CLK_TCK;

    if (msPerTick <= 100) {
        /*
         * On low resolution systems we can do this all with integer math. Note
         * that the addition of half the clock hertz results in appoximate
         * rounding instead of truncation.
         */
        return (numTicks) * (1000 + msPerTick / 2) / msPerTick;
    } else {
        /*
         * On systems (Cray) where the question is ticks per millisecond, not
         * milliseconds per tick, we need to use floating point arithmetic.
         */
        return ((numTicks) * 1000.0 / msPerTick);
    }
}

/*-----------------------------------------------------------------------------
 * TclXOSElapsedTime --
 *   System dependent interface to get the elapsed CPU and real time. 
 *
 * Parameters:
 *   o realTime - Elapsed real time, in milliseconds is returned here.
 *   o cpuTime - Elapsed CPU time, in milliseconds is returned here.
 *-----------------------------------------------------------------------------
 */
void
TclXOSElapsedTime (realTime, cpuTime)
    clock_t *realTime;
    clock_t *cpuTime;
{
/*
 * If times returns elapsed real time, this is easy.  If it returns a status,
 * real time must be obtained in other ways.
 */
#ifndef TIMES_RETS_STATUS
    struct tms cpuTimes;

    *realTime = TclXOSTicksToMS (times (&cpuTimes));
    *cpuTime = TclXOSTicksToMS (cpuTimes.tms_utime + cpuTimes.tms_stime);
#else
    static struct timeval startTime = {0, 0};
    struct timeval currentTime;
    struct tms cpuTimes;

    /*
     * If this is the first call, get base time.
     */
    if ((startTime.tv_sec == 0) && (startTime.tv_usec == 0))
        gettimeofday (&startTime, NULL);
    
    gettimeofday (&currentTime, NULL);
    currentTime.tv_sec  = currentTime.tv_sec  - startTime.tv_sec;
    currentTime.tv_usec = currentTime.tv_usec - startTime.tv_usec;
    *realTime = (currentTime.tv_sec  * 1000) + (currentTime.tv_usec / 1000);
    times (&cpuTimes);
    *cpuTime = TclXOSTicksToMS (cpuTimes.tms_utime + cpuTimes.tms_stime);
#endif
}
