/*
------------------------------------------
prof.c - Profiling [incr Tcl] applications
------------------------------------------

Copyright (C) 1997, 1998 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.9

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h> /* malloc */
#include <limits.h> /* INT_MAX */
#include <string.h> /* strstr */

/* we need access to internal Tcl structures */
#include <tclInt.h>

#include "juf.h"
#include "juftime.h"

/*
 * profiling data 
 */

typedef struct jufprof_procdata
{
	unsigned long  calls;               /* how often invoked */
	jufTimeEntry_t *timeinfos;          /* time information */ 
} jufprof_procdata;

typedef struct jufprof_data
{
	Tcl_Obj*       handle;              /* our profile handle */
	Tcl_HashTable  procdata;            /* per-proc information */
	Tcl_Trace      traceToken;          /* trace identifier */
	int            active;              /* profile active ? */
	Tcl_Obj*       before;              /* script to run before user command */
	Tcl_Obj*       after;               /* script to run after user command */
	jufTimeEntry_t *curtimer;           /* current timer */
} jufprof_data;

/*
 * profiling state information
 */

typedef struct jufprof_state {
	Command        *cmddata;            /* command information */
	Tcl_ObjCmdProc *userobjproc;        /* command function with object ifs */
	Tcl_CmdProc    *userproc;           /* command function */
	ClientData     fdata;               /* command data */
	jufprof_procdata *procdata;         /* profiling data for this command */
	jufprof_data   *profdata;           /* general profiling data */
} jufprof_state;

static Tcl_ObjType *jufprof_type;       /* corresponding Tcl object type */
static Tcl_HashTable jufprof_traces;    /* maps trace handles to data */
static jufprof_handle = 0;              /* profile handle counter */

/* first invocation of `jufprof_hook' sets this to the current state
   information and `jufprof_runobjcmd' resp. `jufprof_runcmd' resets
   it before running the intercepted command */

static jufprof_state *jufprof_state_installed = NULL;

/* protect us from recursions while running `before' and `after' scripts */

static int script_running = 0;

/* Prototypes */
/* ========== */

int jufprof_lookup (Tcl_Interp *interp, Tcl_Obj *objptr);
void jufprof_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj);
void jufprof_tostring (Tcl_Obj *obj);
void jufprof_hook (ClientData clientData, Tcl_Interp *interp, int level,
				   char *command, Tcl_CmdProc *cmdProc,
				   ClientData cmdClientData, int argc, char **argv);
int jufprof_runobjcmd (ClientData fdata, Tcl_Interp *interp, int objc,
					   Tcl_Obj *CONST objv[]);
int jufprof_runcmd (ClientData fdata, Tcl_Interp *interp, int argc,
				   char *argv[]);
int jufprof_runcmd_horse (ClientData fdata, Tcl_Interp *interp,
						  int isObjProc, int argc, void *argptr);

/* Functions */
/* ========= */
	 
/*
 * -------------------------------------------------------------------------
 * jufprof_lookup OBJ
 *
 * Searches for profile represented by OBJ.
 * -------------------------------------------------------------------------
 */

int jufprof_lookup (Tcl_Interp *interp, Tcl_Obj *objptr)
{
	Tcl_HashEntry *entry;
	Tcl_Obj *buf;

	entry = Tcl_FindHashEntry (&jufprof_traces,
							   Tcl_GetStringFromObj (objptr, NULL));
	if (entry == NULL)
		{
		buf = Tcl_NewStringObj ("", 0);
		Tcl_AppendStringsToObj (buf, "no such trace \"", 
								Tcl_GetStringFromObj (objptr, NULL),
								"\"", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	else
		{
		if (objptr -> typePtr -> freeIntRepProc != NULL)
			objptr -> typePtr -> freeIntRepProc (objptr);
		objptr -> internalRep.otherValuePtr = Tcl_GetHashValue (entry);
		objptr -> typePtr = jufprof_type;
		return (TCL_OK);
		}
}

/*
 * ----------------------------------------------------
 * jufprof_dup FROMOBJ TOOBJ
 *
 * Duplicates FROMOBJ and stores the result into TOOBJ.
 * ----------------------------------------------------
 */

void jufprof_dup (Tcl_Obj *fromobj, Tcl_Obj *toobj)
{
	toobj -> internalRep.otherValuePtr = fromobj -> internalRep.otherValuePtr;
}

/*
 * -----------------------------------------------------
 * jufprof_tostring OBJ
 *
 * Creates string representation for OBJ.
 * -----------------------------------------------------
 */

void jufprof_tostring (Tcl_Obj *obj)
{
	 char *buf;
	 jufprof_data *data = (jufprof_data *) obj -> internalRep.otherValuePtr;

	 buf = Tcl_GetStringFromObj (data -> handle, &obj -> length);
	 obj -> bytes = Tcl_Alloc (((unsigned int) obj -> length) + 1);
	 strcpy (obj -> bytes, buf);
}

/*
 * --------------------------------------------------------
 * FUNCTION: jufprof_new_horse
 *
 * Auxiliary function preparing profiling.
 * --------------------------------------------------------
 */

Tcl_Obj *jufprof_new_horse (void)
{
	jufprof_data *data;
    char buf[100];
	Tcl_Obj *prof_new;
	
	/* prepare data structure */
	data = (jufprof_data *) Tcl_Alloc (sizeof (jufprof_data));
	data -> active = 0;         /* tracing disabled */
	data -> before = NULL;      /* no scripts */
	data -> after = NULL;
	data -> curtimer = NULL;    /* no timer active */
	Tcl_InitHashTable (&data -> procdata, TCL_STRING_KEYS);
	sprintf (buf, "trace%i", jufprof_handle++);

	/* create new object */
	prof_new = Tcl_NewObj ();
	prof_new -> internalRep.otherValuePtr = data;
	prof_new -> typePtr = jufprof_type;
	Tcl_IncrRefCount (prof_new);

	return (prof_new);
}

/*
 * --------------------------------------------------------
 * FUNCTION: jufprof_new
 *
 * Prepares profiling and returns handle.
 * --------------------------------------------------------
 */

int jufprof_new (ClientData fdata, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[])
{
	/* check argument number */
	if (objc > 1)
		{
		Tcl_AppendResult (interp, "wrong # args: should be ",
						  Tcl_GetStringFromObj (objv[0], NULL),
						  "", NULL);
		return(TCL_ERROR);
		}

	
	/* set result to new object */
	Tcl_SetObjResult (interp, jufprof_new_horse ());
	return (TCL_OK);
}

/*
 * --------------------------------------------------------
 * COMMAND: jufprof_start [HANDLE]
 *
 * Starts profiling and returns handle.
 * --------------------------------------------------------
 */

int jufprof_start (ClientData fdata, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[])
{
	jufprof_data *data;
	Tcl_Obj *profobj;
	
	/* check argument number */
	if (objc > 2)
		{
		Tcl_AppendResult (interp, "wrong # args: should be ",
						  Tcl_GetStringFromObj (objv[0], NULL),
						  " ?profHandle?", NULL);
		return(TCL_ERROR);
		}
	else if (objc == 2)
		{
		/* reusing profile handle */
		if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
			{
			/* not a valid profile representation */
			return (TCL_ERROR);
			}
		
		profobj = objv[1];
		}
	else
		{
		/* create profile handle */
		profobj = jufprof_new_horse ();
		}

	data = profobj -> internalRep.otherValuePtr;

	/* start tracing */
	data -> traceToken = Tcl_CreateTrace (interp, INT_MAX, jufprof_hook, data);
	data -> active = 1;
	
	/* set result to new object */
	Tcl_SetObjResult (interp, profobj);
	return (TCL_OK);
}

/*
 * ------------------------------------------
 * COMMAND: jufprof_stop OBJ
 *
 * Stops profiling.
 * ------------------------------------------
 */

int jufprof_stop (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	jufprof_data *data;
  
	/* check argument number */
	if (objc != 2)
		{
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;

	/* stop tracing */
	if (data -> active)
		{
		Tcl_DeleteTrace (interp, data -> traceToken);
		data -> active = 0;
		data -> curtimer = NULL;
		}
	else
		{
		Tcl_AppendStringsToObj (buf, "profiling not active for ",
								Tcl_GetStringFromObj (objv[1], NULL),
								NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	return (TCL_OK);
}

/*
 * ------------------------------------------
 * COMMAND: jufprof_delete OBJ
 *
 * Deletes profiling object.
 * ------------------------------------------
 */

int jufprof_delete (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
  
	/* check argument number */
	if (objc != 2)
		{
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	Tcl_DecrRefCount (objv[1]);
	return (TCL_OK);
}

/*
 * ----------------------------------------------------------------
 * COMMAND: jufprof_commands HANDLE
 *
 * Returns a list of all commands called while profiling was active
 * for HANDLE.
 * ----------------------------------------------------------------
 */

int jufprof_commands (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	Tcl_Obj *list;
	jufprof_data *data; char *command;
	Tcl_HashEntry *entry; Tcl_HashSearch hsrch;
	
	/* check argument number */
	if (objc != 2)
		{
		Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
		
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;
	list = Tcl_NewListObj (0, NULL);
	
	/* loop over hash entries */
	entry = Tcl_FirstHashEntry (& (data->procdata), &hsrch);
	while (entry != NULL)
		{
		command = Tcl_GetHashKey (& (data->procdata), entry);
		Tcl_ListObjAppendElement (interp, list,
								  Tcl_NewStringObj (command,
													(int) strlen(command)));
		entry = Tcl_NextHashEntry (&hsrch);
		}

	Tcl_SetObjResult (interp, list);		
	return (TCL_OK);
}

/*
 * ----------------------------------------------------------------
 * COMMAND: jufprof_calls HANDLE NAME
 *
 * Returns number of calls to command NAME while profiling for
 * HANDLE was active.
 * ----------------------------------------------------------------
 */

int jufprof_calls (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	jufprof_data *data;
	jufprof_procdata *procdata;
	Tcl_HashEntry *entry;
	
	/* check argument number */
	if (objc != 3)
		{
		Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
		
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle command", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;
	
	/* lookup date for COMMAND */
	entry = Tcl_FindHashEntry (& (data->procdata),
							   Tcl_GetStringFromObj (objv[2], NULL));
	if (entry != NULL)
		{
		/* COMMAND has been called at least once while we were watching */
		procdata = Tcl_GetHashValue (entry);
		Tcl_SetObjResult (interp, Tcl_NewLongObj ((long) procdata -> calls));
		}
	else
		{
		Tcl_SetObjResult (interp, Tcl_NewLongObj (0));
		}

	return (TCL_OK);
}

/*
 * ----------------------------------------------------------------
 * COMMAND: jufprof_elapsed HANDLE NAME
 *
 * Returns time information for command NAME.
 * ----------------------------------------------------------------
 */

int jufprof_elapsed (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	jufprof_data *data;
	jufprof_procdata *procdata;
	Tcl_HashEntry *entry;
	
	/* check argument number */
	if (objc != 3)
		{
		Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
		
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle command", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;
	
	/* lookup date for COMMAND */
	entry = Tcl_FindHashEntry (& (data->procdata),
							   Tcl_GetStringFromObj (objv[2], NULL));
	if (entry != NULL)
		{
		/* COMMAND has been called at least once while we were watching */
		procdata = Tcl_GetHashValue (entry);
		Tcl_SetObjResult (interp, Tcl_NewLongObj ((long) procdata -> timeinfos -> real_sum));
		}
	else
		{
		Tcl_SetObjResult (interp, Tcl_NewLongObj (0));
		}

	return (TCL_OK);
}

/*
 * ----------------------------------------------------------------
 * COMMAND: jufprof_after HANDLE SCRIPT
 *
 * Registers script to be executed after a user command is
 * executed when tracing is active for HANDLE.
 * ----------------------------------------------------------------
 */

int jufprof_after (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	jufprof_data *data;
	
	/* check argument number */
	if (objc != 3)
		{
		Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
		
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle script", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;

	/* register script */
	if (data -> after != NULL)
		{
		/* remove old script */
		Tcl_DecrRefCount (data -> after);
		data -> after = NULL;
		}

	data -> after = objv[2];
	return (TCL_OK);
}

/*
 * ----------------------------------------------------------------
 * COMMAND: jufprof_before HANDLE SCRIPT
 *
 * Registers script to be executed before a user command is
 * executed when tracing is active for HANDLE.
 * ----------------------------------------------------------------
 */

int jufprof_before (ClientData fdata, Tcl_Interp *interp,  int objc,
				Tcl_Obj *CONST objv[])
{
	jufprof_data *data;
	
	/* check argument number */
	if (objc != 3)
		{
		Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
		
		Tcl_AppendStringsToObj (buf, "wrong # args: should be ",
					  Tcl_GetStringFromObj (objv[0], NULL),
					  " profHandle script", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}  

	if (Tcl_ConvertToType (interp, objv[1], jufprof_type) == TCL_ERROR)
		{
		/* not a valid profile representation */
		return (TCL_ERROR);
		}

	data = objv[1] -> internalRep.otherValuePtr;

	/* register script */
	if (data -> before != NULL)
		{
		/* remove old script */
		Tcl_DecrRefCount (data -> before);
		data -> before = NULL;
		}

	data -> before = objv[2];
	return (TCL_OK);
}

/*
 * -------------------------------------------------
 * FUNCTION: jufprof_hook
 *
 * Trace handler.
 * -------------------------------------------------
 */

void jufprof_hook (ClientData clientData, Tcl_Interp *interp, int level,
				   char *command, Tcl_CmdProc *cmdProc,
				   ClientData cmdClientData, int argc, char **argv)
{
	Tcl_CmdInfo cmdinfo; Tcl_Obj *buf;
	jufprof_data *data = clientData;
	jufprof_procdata *procdata; char *name, *cmd, *pos;
	jufprof_state *state;
	Tcl_HashEntry *pEntry; int firstcall;
	Command *cmddata;
	
	/* check if before/after script is running */
	if (script_running)
		return;

	/* suspend timer if necessary */
	if (data -> curtimer != NULL)
		JufTimerSuspend (data -> curtimer);
	
	/* build fully qualified command name */
	if (! Tcl_GetCommandInfo (interp, argv[0], &cmdinfo))
		{
		fprintf (stderr, "unknown command \"%s\" (complete command: %s)\n",
				 argv[0], command);
		return;
		}

	if (strncmp (argv[0], "::", 2) == 0)
		{
		/* command already fully qualified */
		name = argv[0] + 2;
		buf = Tcl_NewStringObj (name, (int) strlen (name));
		}
	else if (strlen (cmdinfo.namespacePtr -> name))
		{
		/* add namespace containing the command */
		name = cmdinfo.namespacePtr -> fullName + 2;
		buf = Tcl_NewStringObj (name, (int) strlen (name));
		/* skip namespaces within command name */
		cmd = argv[0];
		while ((pos = strstr (cmd, "::")) != NULL)
			cmd = pos += 2;
		Tcl_AppendStringsToObj (buf, "::", cmd, NULL);
		}
	else 
		{
		/* toplevel command */
		buf = Tcl_NewStringObj (argv[0], (int) strlen (argv[0]));
		}
								
	/* count function call */
	pEntry = Tcl_CreateHashEntry (&data -> procdata,
								  Tcl_GetStringFromObj (buf, NULL),
								  &firstcall);
	if (firstcall)
		{
		procdata = (jufprof_procdata *) Tcl_Alloc
			((sizeof (jufprof_procdata)));
		procdata -> calls = 1;
		procdata -> timeinfos = JufTimerAlloc ();
		Tcl_SetHashValue (pEntry, procdata);
		}
	else
		{
		procdata = Tcl_GetHashValue (pEntry);
		procdata -> calls++;
		}

	/* install not more than once for a given command */
	if (jufprof_state_installed == NULL)
		{
		/* get a hand on the internal structure */
		cmddata = (Command *) Tcl_FindCommand
			(interp, Tcl_GetStringFromObj (buf, NULL),
			 NULL, TCL_LEAVE_ERR_MSG);
		
		if (cmddata == NULL)
			{
			fprintf (stderr, "command data not found: %s\n",
					 Tcl_GetStringResult (interp));
			exit (1);
			}

		/* create state information */
		state = (jufprof_state *) Tcl_Alloc (sizeof (jufprof_state));
		state -> cmddata = cmddata;
		state -> userobjproc = NULL;
		state -> userproc = NULL;
		state -> fdata = cmdClientData;
		state -> procdata = procdata;
		state -> profdata = data;

		/* set current state information */
		jufprof_state_installed = state;

		/* replace user command with wrapper function */
		if (cmdinfo.isNativeObjectProc)
			{
			state -> userobjproc = cmddata -> objProc;
			cmddata -> objProc = jufprof_runobjcmd;
			}
		else
			{
			state -> userproc = cmddata -> proc;
			cmddata -> proc = jufprof_runcmd;
			}
		}
}

/*
 * --------------------------------------------------------------
 * FUNCTION: jufprof_runobjcmd
 *
 * Wrapper around user commands. Does time recording and
 * other vital profiling stuff.
 * --------------------------------------------------------------
 */

int jufprof_runobjcmd (ClientData fdata, Tcl_Interp *interp, int objc,
				   Tcl_Obj *CONST objv[])
{
	return (jufprof_runcmd_horse (fdata, interp, 1, objc, (void *) objv));
}

/*
 * --------------------------------------------------------------
 * FUNCTION: jufprof_runcmd
 *
 * Wrapper around user commands. Does time recording and
 * other vital profiling stuff.
 * --------------------------------------------------------------
 */

int jufprof_runcmd (ClientData fdata, Tcl_Interp *interp, int argc,
				   char *argv[])
{
	return (jufprof_runcmd_horse (fdata, interp, 0, argc, argv));
}

/*
 * ----------------------------------------------------------------
 * FUNCTION: jufprof_runcmd_horse
 *
 * Does the real work for `jufprof_runobjcmd' and `jufprof_runcmd'.
 * ISOBJPROC indicates if called by the former (<> 0)
 * or the latter (= 0).
 * ----------------------------------------------------------------
 */

int jufprof_runcmd_horse (ClientData fdata, Tcl_Interp *interp,
						  int isObjProc, int argc, void *argptr)
{
  jufprof_state *state;
  int result;
  jufTimeEntry_t *curtimer;
  Tcl_Obj *resultobj;
  
  /* Move global state information to local stack */
  state = jufprof_state_installed;
  jufprof_state_installed = NULL;

  /* Copy time information for last command to local stack */
  curtimer = state -> profdata -> curtimer;
 
  /* Restore user command */
  if (isObjProc)
	  state -> cmddata -> objProc = state -> userobjproc;
  else
	  state -> cmddata -> proc = state -> userproc;
  
  /* Run before hook */
  if (state -> profdata -> before != NULL)
	  {
	  script_running = 1;
	  Tcl_EvalObj (interp, state -> profdata -> before);
	  Tcl_ResetResult (interp);   /* discard result */
	  script_running = 0;
	  }
  
  /* Call the command we intercepted */
  state -> profdata -> curtimer = state -> procdata -> timeinfos;
  JufTimerCont (state -> procdata -> timeinfos);

  if (isObjProc)
	  result = (*state->userobjproc) (fdata, interp, argc,
									  (Tcl_Obj * CONST *) argptr);
  else
	  result = (*state->userproc) (fdata, interp, argc, (char **) argptr);
  
  JufTimerSuspend (state -> procdata -> timeinfos);
  state -> profdata -> curtimer = state -> procdata -> timeinfos;
  
  /* Run after hook */
  if (state -> profdata -> after != NULL)
	  {
	  script_running = 1;

	  /* Record protected result of user command */
	  resultobj = Tcl_DuplicateObj (Tcl_GetObjResult (interp));
	  
	  Tcl_EvalObj (interp, state -> profdata -> after);

	  /* Restore result of user command */

	  Tcl_SetObjResult (interp, resultobj);
	  
	  script_running = 0;
	  }

  /* Restore timer */
  if (curtimer != NULL)
	  JufTimerCont (curtimer);

  return (result);
}

/*
 * ---------------------------------------------------
 * FUNCTION: Jufprof_Init INTERP
 *
 * Initializes profiling package.
 * ---------------------------------------------------
 */
	
int
Jufprof_Init (Tcl_Interp *interp)
{
	if (jufprof_type == NULL)
		{
		/* create new Tcl object type */
		jufprof_type = (Tcl_ObjType *) Tcl_Alloc (sizeof (Tcl_ObjType));
		jufprof_type -> name = "Profiledata";
		jufprof_type -> freeIntRepProc = NULL;
		jufprof_type -> dupIntRepProc =  jufprof_dup;
		}

	/* create Tcl commands */
	Tcl_CreateObjCommand (interp, "Juf::Prof::new", jufprof_new,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::start", jufprof_start,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::stop", jufprof_stop,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::commands", jufprof_commands,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::calls", jufprof_calls,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::elapsed", jufprof_elapsed,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::delete", jufprof_delete,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::before", jufprof_before,
						  NULL, NULL);
	Tcl_CreateObjCommand (interp, "Juf::Prof::after", jufprof_after,
						  NULL, NULL);
	
	/* provide JUFPROF as package */
	return (Tcl_PkgProvide (interp, JUFPROF_PKGNAME, JUFPROF_VERSION));
}

