/*
-------------------------------------------------------  
rpm.c - Interface to the RPM library
-------------------------------------------------------

Copyright (C) 1998 Stefan Hornburg

Author: Stefan Hornburg <racke@gundel.han.de>
Maintainer: Stefan Hornburg <racke@gundel.han.de>
Version: 0.0.9

This file is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

This file is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this file; see the file COPYING.  If not, write to the Free
Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* strdup */
#include <ctype.h> /* tolower */
#include <errno.h>
#include <fcntl.h>  /* Unix file ops */
#include <unistd.h> /* ditto */

#include <tcl.h>
#include <rpm/rpmlib.h>

#include "juf.h"

/* Data structures */
/* =============== */

typedef struct jufrpm_errinfo
{
	struct jufrpm_errinfo *last;    /* previous error informations */
	Tcl_Interp *interp;             /* active Tcl interpreter */
	int code;                       /* RPM error code */
	char *string;                   /* RPM error string (allocated !) */
} jufrpm_errinfo;

/* maps strings for `query' command to RPM tags */
typedef struct jufrpm_tagmap
{
	int_32 tag;                     /* RPM tag */
	char *name;                     /* corresponding name */
} jufrpm_tagmap;

/* Variables */
/* ========= */

/* maps strings for `query' command to RPM tags */
static jufrpm_tagmap tagmap [] = {
	{RPMTAG_DESCRIPTION, "description"},
	{RPMTAG_FILENAMES, "files"},
	{RPMTAG_GROUP, "group"},
	{RPMTAG_NAME, "name"},
	{RPMTAG_OBSOLETES, "obsoletes"},
	{RPMTAG_RELEASE, "release"},
	{RPMTAG_SOURCERPM, "sourcerpm"},
	{RPMTAG_SUMMARY, "summary"},
	{RPMTAG_URL, "url"},
	{RPMTAG_VERSION, "version"},
};

static int tagmapsize = sizeof(tagmap) / sizeof(tagmap[0]);

/* hash table with query tags */
static Tcl_HashTable jufrpm_tags;
/* hash table with RPM variables */
static Tcl_HashTable jufrpm_vars;
 /* RPM error informations */
static jufrpm_errinfo *jufrpm_errinfoptr = NULL;
 /* active interpreter */
static Tcl_Interp *jufrpm_errinterp = NULL;

/* Prototypes */
/* ========== */

void jufrpm_errhandler (void);
void jufrpm_graberrmsg (Tcl_Interp *interp, Tcl_Obj *obj);
void jufrpm_wrap (Tcl_Interp *interp);

int jufrpm_open (Tcl_Interp *interp, rpmdb *db, int mode);
void jufrpm_close (rpmdb db);
Tcl_Obj *jufrpm_entry (Tcl_Interp *interp, Header h, int_32 tag);
char *jufrpm_var (ClientData data, Tcl_Interp *interp, char *name1,
						char *name2, int flags);

/* prototypes for Tcl commands */
int jufrpm_exists (ClientData data, Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]);
int jufrpm_list (ClientData data, Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]);
int jufrpm_query (ClientData data, Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]);
int jufrpm_install (ClientData data, Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]);

/*
 * ---------------------------------------------------------------
 * FUNCTION: jufrpm_errhandler
 *
 * Handler that receives error notifications from the RPM library.
 * ---------------------------------------------------------------
 */

void jufrpm_errhandler (void)
{
	jufrpm_errinfo *errinfoptr;
	
	if (jufrpm_errinterp == NULL)
		{
		fprintf (stderr, "RPM error infos: %s (%i)\n",
				 rpmErrorString (), rpmErrorCode ());
		fprintf (stderr, "inactive RPM error handler notified\n");
		exit (1);
		}

	/* store error notification */
	errinfoptr = (jufrpm_errinfo *) Tcl_Alloc (sizeof (jufrpm_errinfo));
	errinfoptr -> last = jufrpm_errinfoptr;
	errinfoptr -> code = rpmErrorCode ();
	errinfoptr -> string = Tcl_Alloc (strlen (rpmErrorString () + 1));
	strcpy (errinfoptr -> string, rpmErrorString ());
	errinfoptr -> interp = jufrpm_errinterp;
	jufrpm_errinfoptr = errinfoptr;	
}

/*
 * --------------------------------------------------
 * FUNCTION: jufrpm_wrap INTERP
 *
 * (De)activates error handler for RPM notifications.
 * --------------------------------------------------
 */

void jufrpm_wrap (Tcl_Interp *interp)
{
	if (interp == NULL)
		{
		if (jufrpm_errinterp == NULL)
			{
			fprintf (stderr,
					 "couldn't deactivate error handler for RPM notifications\n");
			exit (1);
			}
		jufrpm_errinterp = NULL;
		}
	else if (jufrpm_errinterp != NULL)
		{
		fprintf (stderr, "error handler for RPM notfications already activated\n");
		exit (1);
		}
	jufrpm_errinterp = interp;
}

/*
 * ----------------------------------------------------
 * FUNCTION: jufrpm_graberrmsg INTERP OBJ
 *
 * Removes and records recent error message within OBJ.
 * ----------------------------------------------------
 */

void jufrpm_graberrmsg (Tcl_Interp *interp, Tcl_Obj *obj)
{
	jufrpm_errinfo *last;
	
	if (jufrpm_errinfoptr != NULL)
		{
		Tcl_AppendStringsToObj (obj, jufrpm_errinfoptr -> string,
								NULL);
		last = jufrpm_errinfoptr -> last;
		Tcl_Free (jufrpm_errinfoptr -> string);
		Tcl_Free ((char *) jufrpm_errinfoptr);
		jufrpm_errinfoptr = last;
		}
	else
		{
		Tcl_AppendStringsToObj (obj,
								"no error information available from RPM",
								NULL);
		}
}

/*
 * -----------------------------------------------------------------
 * FUNCTION: jufrpm_open INTERP DB MODE
 *
 * Opens database and stores data into DB. Returns Tcl return value.
 * -----------------------------------------------------------------
 */

int jufrpm_open (Tcl_Interp *interp, rpmdb *db, int mode)
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	int status;
	
	/* determine database location */
	if (rpmGetVar (RPMVAR_DBPATH) == NULL)
		{
		Tcl_AppendStringsToObj (buf, "couldn't locate RPM database", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	
	/* open database */
	jufrpm_wrap (interp);
	status = rpmdbOpen ("", db, mode, 0);
	jufrpm_wrap (NULL);

	if (status)
		{
		Tcl_AppendStringsToObj (buf, "couldn't open RPM database: ",
								strerror (errno), NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	return (TCL_OK);
}

/*
 * -----------------------------------------------------------------
 * FUNCTION: jufrpm_close DB
 *
 * Closes database and frees any data allocated by DB.
 * -----------------------------------------------------------------
 */

void jufrpm_close (rpmdb db)
{
	rpmdbClose (db);
}

/*
 * ------------------------------------------------------
 * FUNCTION: jufrpm_entry INTERP HDR TAG
 *
 * Reads entry from package header HDR for entry matching
 * TAG and returns result if sucessful, NULL otherwise.
 * ------------------------------------------------------
 */

Tcl_Obj *jufrpm_entry (Tcl_Interp* interp, Header hdr, int_32 tag)
{
	int_32 type; void *data; char **strings; int_32 count;
	Tcl_Obj *list, *buf; int_32 i;
	
	/* reads package entry */
	if (headerGetEntry (hdr, tag, &type, &data, &count))
		{
		/* successful */
		switch (type)
			{
			case RPM_STRING_TYPE:
				return (Tcl_NewStringObj (data, (int) strlen (data)));
			case RPM_STRING_ARRAY_TYPE:
				/* build a list from string array */
				list = Tcl_NewListObj (0, NULL);
				strings = (char **) data;
				for (i = 0; i < count; i++)
					{
					buf = Tcl_NewStringObj (strings[i],
											(int) strlen (strings[i]));
					Tcl_ListObjAppendElement (interp, list, buf);
					}

				free (data);
				return (list);
			default:
				return (NULL);
			}
		}
	else
		{
		return (NULL);
		}
}

/*
 * --------------------------------------------------
 * FUNCTION: jufrpm_var DATA INTERP NAME1 NAME2 FLAGS
 *
 * Link between Tcl variables and RPM variables.
 * --------------------------------------------------
 */

char *jufrpm_var (ClientData data, Tcl_Interp *interp, char *name1,
				  char *name2, int flags)
{
	int symbol = (int) data;
	char *value;
	
	if (flags & TCL_TRACE_WRITES)
		{
		value = Tcl_GetVar2 (interp, name1, name2, 0);
		rpmSetVar (symbol, value);
		}
	else if (flags & TCL_TRACE_READS)
		{
		if ((value = rpmGetVar (symbol)) == NULL)
			value = "";
		Tcl_SetVar2 (interp, name1, name2, value, 0);
		}

	return (NULL);
}

/*
 * -----------------------------------------
 * COMMAND: jufrpm_exists PACKAGE
 *
 * Returns 1 if PACKAGE exists, 0 otherwise.
 * -----------------------------------------
 */

int jufrpm_exists (ClientData data, Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	rpmdb db; int state; dbiIndexSet set;
  
	/* check argument number */
	if (objc != 2)
		{
		Tcl_AppendStringsToObj (buf, "wrong #args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" packageName", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	/* open database */
	if (jufrpm_open (interp, &db, O_RDONLY) == TCL_ERROR)
		return (TCL_ERROR);

	/* look for match */
	state = rpmdbFindPackage (db, Tcl_GetStringFromObj (objv[1], NULL), &set);
	jufrpm_close (db);
	
	if (state == -1)
		{
		/* read error */
		Tcl_AppendStringsToObj (buf, "search in RPM database \"",
								rpmGetVar (RPMVAR_DBPATH), "\" failed: ",
								strerror (errno), NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}
	else if (state == 0)
		{
		/* package found */
		Tcl_SetObjResult (interp, Tcl_NewIntObj (1));
		}
	else
		Tcl_SetObjResult (interp, Tcl_NewIntObj (0));
		
	return (TCL_OK);
}

/*
 * -----------------------------------------
 * COMMAND: jufrpm_list
 *
 * Returns list with all installed packages.
 * -----------------------------------------
 */

int jufrpm_list (ClientData data, Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	Tcl_Obj *list = Tcl_NewListObj (0, NULL);
	
	rpmdb db; unsigned int no; Header pkghdr;
	char *name;
	int_32 type; int count;
  
	/* check argument number */
	if (objc != 1)
		{
		Tcl_AppendStringsToObj (buf, "wrong #args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								"", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	/* open database */
	if (jufrpm_open (interp, &db, O_RDONLY) == TCL_ERROR)
		return (TCL_ERROR);

	/* traverse database */
	no = rpmdbFirstRecNum (db);
	while (no != 0)
		{
		/* read header */
		if ((pkghdr = rpmdbGetRecord (db, no)) == NULL)
			{
			Tcl_AppendStringsToObj (buf, "couldn't read package information",
									NULL);
			Tcl_SetObjResult (interp, buf);
			}

		/* add package name to list */
		headerGetEntry (pkghdr, RPMTAG_NAME, &type, (void **) &name, &count);
		Tcl_ListObjAppendElement (interp, list,
								  Tcl_NewStringObj (name,
													(int) strlen (name)));
		no = rpmdbNextRecNum (db, no);
		}
	
	/* close database */
	jufrpm_close (db);
	Tcl_SetObjResult (interp, list);
	return (TCL_OK);
}

/*
 * -------------------------------------------------------
 * COMMAND: jufrpm_query PACKAGE [TAG]
 *
 * Queries PACKAGE. If TAG is given, return value for it.
 * Otherwise return list with tags and values as elements.
 * -------------------------------------------------------
 */

int jufrpm_query (ClientData data, Tcl_Interp *interp, int objc,
				  Tcl_Obj *CONST objv[]) {
  Tcl_Obj *buf = Tcl_NewStringObj ("", 0), *entrybuf;
  char *package, *tag = NULL;
  int fd, state = 1;
  Header pkghdr; rpmdb db; dbiIndexSet set;
  Tcl_HashEntry *entry;
  
  /* check argument number */
  if (objc != 2 && objc != 3) {
	Tcl_AppendStringsToObj (buf, "wrong #args: should be ",
							Tcl_GetStringFromObj (objv[0], NULL),
							" package ?tag?", NULL);
	Tcl_SetObjResult (interp, buf);
	return (TCL_ERROR);
  }

  /* prepare arguments */
  package = Tcl_GetStringFromObj (objv[1], NULL);
  if (objc == 3)
	  tag = Tcl_GetStringFromObj (objv[2], NULL);

  /* use file as source for package info if one exists */
  if ((fd = open (package, O_RDONLY)) != -1) {
	state = rpmReadPackageInfo (fd, NULL, &pkghdr);
	close (fd);
	if (state) {
	  /* failure */
	  Tcl_AppendStringsToObj (buf, "couldn't read package header from \"",
							  package, "\": ", strerror (errno), NULL);
	  Tcl_SetObjResult (interp, buf);
	  return (TCL_ERROR);
	}
  } else {
	/* check if we have to bail out or try the database */
	switch (errno) {
	case EISDIR: /* file is a directory */
	case ENOENT: /* file doesn't exist */
	  break;
	default:
	  Tcl_AppendStringsToObj (buf, "couldn't open \"", package, "\": ",
							  strerror (errno), NULL);
	  Tcl_SetObjResult (interp, buf);
	  return (TCL_ERROR);
	}
  }

  if (state)
	  {
	  /* open database */
	  if (jufrpm_open (interp, &db, O_RDONLY) == TCL_ERROR)
		  return (TCL_ERROR);
	  
	  /* look for matches */
	  if ((state = rpmdbFindPackage (db, package, &set)))
		  {
		  if (state == 1)
			  {
			  /* no match */
			  jufrpm_close (db);
			  Tcl_AppendStringsToObj (buf, "package \"",
									  package, "\" not found in RPM database",
									  NULL);
			  Tcl_SetObjResult (interp, buf);
			  return (TCL_ERROR);
			  }
		  else
			  {
			  /* read error */
			  jufrpm_close (db);
			  Tcl_AppendStringsToObj (buf, "search in RPM database \"",
									  rpmGetVar (RPMVAR_DBPATH), "\" failed: ",
									  strerror (errno), NULL);
			  Tcl_SetObjResult (interp, buf);
			  return (TCL_ERROR);
			  }
		  }
	  else
		  {
		  if (set.count == 1)
			  {
			  /* read header */
			  if ((pkghdr = rpmdbGetRecord (db, set.recs[0].recOffset))
				  == NULL)
				  {
				  dbiFreeIndexRecord (set);
				  jufrpm_close (db);
				  Tcl_AppendStringsToObj (buf, "couldn't read header for package \"",
										  package, "\": ",
										  strerror (errno), NULL);
				  Tcl_SetObjResult (interp, buf);
				  return (TCL_ERROR);		  
				  }
			  }
		  else
			  {
			  dbiFreeIndexRecord (set);
			  jufrpm_close (db);
			  Tcl_AppendStringsToObj (buf, "multiple packages matching \"",
									  package, "\" found in RPM database",
								NULL);
			  Tcl_SetObjResult (interp, buf);
			  return (TCL_ERROR);
			  }
		  }
	  }

  /* header successfully read, now fetch information */
  if (objc == 2)
	  {
	  Tcl_HashSearch hsrch; char *tagname;
	  Tcl_Obj *listobj = Tcl_NewListObj (0, NULL);
	  
	  /* build mapping with tag/value pairs */
	  entry = Tcl_FirstHashEntry (&jufrpm_tags, &hsrch);
	  while (entry != NULL)
		  {
		  tagname = Tcl_GetHashKey (&jufrpm_tags, entry);
		  entrybuf = Tcl_NewStringObj (tagname, (int) strlen (tagname));
		  
		  /* append tag name */
		  Tcl_ListObjAppendElement (interp, listobj, entrybuf);
		  /* append tag value */
		  entrybuf = jufrpm_entry (interp, pkghdr,
								   (int) Tcl_GetHashValue (entry));
		  if (entrybuf == NULL)
			  entrybuf = Tcl_NewStringObj ("", 0);
		  Tcl_ListObjAppendElement (interp, listobj, entrybuf);
		  entry = Tcl_NextHashEntry (&hsrch);
		  }
	  Tcl_SetObjResult (interp, listobj);
	  }
  else
	  {
      /* now check tag */
	  if ((entry = Tcl_FindHashEntry (&jufrpm_tags, tag)) == NULL)
		  {
		  Tcl_AppendStringsToObj (buf, "invalid or unsupported tag \"",
								  tag, "\"", NULL);
		  Tcl_SetObjResult (interp, buf);
		  return (TCL_ERROR);
		  }
  
	  /* read tag from header */
	  entrybuf = jufrpm_entry (interp, pkghdr,
							   (int) Tcl_GetHashValue (entry));
	  if (entrybuf == NULL)
		  entrybuf = Tcl_NewStringObj ("", 0);
	  
	  Tcl_SetObjResult (interp, entrybuf);
	  }
  
  return (TCL_OK);
}

/*
 * -----------------------------------------
 * COMMAND: jufrpm_install PACKAGE
 *
 * Installs PACKAGE.
 * -----------------------------------------
 */

int jufrpm_install (ClientData data, Tcl_Interp *interp, int objc,
				 Tcl_Obj *CONST objv[])
{
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);
	Tcl_Obj *errbuf = Tcl_NewStringObj ("", 0);
	Tcl_Obj *pkgname;
	
	rpmdb db; Header pkghdr;
	int result;
	char *pkgfile;
	int fd;	
	int isSource;
	
	/* check argument number */
	if (objc != 2)
		{
		Tcl_AppendStringsToObj (buf, "wrong #args: should be ",
								Tcl_GetStringFromObj (objv[0], NULL),
								" pkgfile", NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	pkgfile = Tcl_GetStringFromObj (objv[1], NULL);
	
	/* open package file */
	if ((fd = open (pkgfile, O_RDONLY)) == -1)
		{
		Tcl_AppendStringsToObj (buf, "couldn't open \"", pkgfile, "\": ",
								strerror (errno), NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	/* read package information */
	result = rpmReadPackageHeader (fd, &pkghdr, &isSource, NULL, NULL);

	if (result) {
	  /* failure */
	  Tcl_AppendStringsToObj (buf, "couldn't read package header from \"",
							  pkgfile, "\": ", strerror (errno), NULL);
	  Tcl_SetObjResult (interp, buf);
	  close (fd);
	  return (TCL_ERROR);
	}

	/* retrieve package name */
	pkgname = jufrpm_entry (interp, pkghdr, RPMTAG_NAME);
	
	/* open database */
	if (jufrpm_open (interp, &db, O_RDWR) == TCL_ERROR)
		return (TCL_ERROR);

	/* install package */
	if (isSource)
		{
		result = rpmInstallSourcePackage (NULL, fd,
										  NULL,		/* specfile */
										  NULL,		/* notify function */
										  NULL, 	/* label format */
										  NULL);    /* cookies */
		}
	else
		{
		result = rpmInstallPackage (NULL, db,
									fd, NULL, 		/* relocations */
									0, 				/* flags */
									NULL, 			/* notify function */
									NULL); 			/* labelformat */
		}
	
	/* close package file */
	close (fd);
	
	/* close database */
	jufrpm_close (db);

	if (result == 0)
		{
		return (TCL_OK);
		}
	else
		{
		/* retrieve RPM's last error message */
		jufrpm_graberrmsg (interp, errbuf);
		Tcl_AppendStringsToObj (buf, "couldn't install package \"",
								Tcl_GetStringFromObj (pkgname, NULL),
								"\": ",
								Tcl_GetStringFromObj (errbuf, NULL),
								NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);		
		}
}

/*
 * ---------------------------------------------------
 * FUNCTION: Jufrpm_Init
 *
 * Initializes Juf::RPM package. Tries to read RPM
 * configuration files.
 * ---------------------------------------------------
 */

int Jufrpm_Init (Tcl_Interp *interp) {
	int flag; char *name; char *value; int i;
	Tcl_HashEntry *entry; Tcl_HashSearch hsrch;
	Tcl_Obj *buf = Tcl_NewStringObj ("", 0);

	/* install handler for RPM error messages */
	rpmErrorSetCallback (jufrpm_errhandler);
	
	/* permit RPM messages only for fatal errors */
	rpmSetVerbosity (RPMMESS_FATALERROR);
	
	/* setup RPM configuration */
	if (rpmReadConfigFiles (NULL, NULL, NULL, 0))
		{
		Tcl_AppendStringsToObj (buf, "couldn't read RPM configuration: ",
								strerror (errno), NULL);
		Tcl_SetObjResult (interp, buf);
		return (TCL_ERROR);
		}

	/* fill variables hash table */
	Tcl_InitHashTable (&jufrpm_vars, TCL_STRING_KEYS);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "sourcedir", &flag),
					  RPMVAR_SOURCEDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "builddir", &flag),
					  RPMVAR_BUILDDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "optflags", &flag),
					  RPMVAR_OPTFLAGS);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "topdir", &flag),
					  RPMVAR_TOPDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "specdir", &flag),
					  RPMVAR_SPECDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "rpmdir", &flag),
					  RPMVAR_RPMDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "srpmdir", &flag),
					  RPMVAR_SRPMDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "distribution",
										   &flag),
					  RPMVAR_DISTRIBUTION);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "vendor", &flag),
					  RPMVAR_VENDOR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "messagelevel",
										   &flag),
					  RPMVAR_MESSAGELEVEL);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "timecheck", &flag),
					  RPMVAR_TIMECHECK);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "sigtype", &flag),
					  RPMVAR_SIGTYPE);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "pgp_path", &flag),
					  RPMVAR_PGP_PATH);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "pgp_name", &flag),
					  RPMVAR_PGP_NAME);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "excludedocs", &flag),
					  RPMVAR_EXCLUDEDOCS);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "buildroot", &flag),
					  RPMVAR_BUILDROOT);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "dbpath", &flag),
					  RPMVAR_DBPATH);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "packager", &flag),
					  RPMVAR_PACKAGER);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "ftpproxy", &flag),
					  RPMVAR_FTPPROXY);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "tmppath", &flag),
					  RPMVAR_TMPPATH);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "ftpport", &flag),
					  RPMVAR_FTPPORT);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "netsharedpath",
										   &flag),
					  RPMVAR_NETSHAREDPATH);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "defaultdocdir",
										   &flag),
					  RPMVAR_DEFAULTDOCDIR);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "fixperms", &flag),
					  RPMVAR_FIXPERMS);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "gzipbin", &flag),
					  RPMVAR_GZIPBIN);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "rpmfilename", &flag),
					  RPMVAR_RPMFILENAME);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "provides", &flag),
					  RPMVAR_PROVIDES);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "buildshell", &flag),
					  RPMVAR_BUILDSHELL);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "instchangelog",
										   &flag),
					  RPMVAR_INSTCHANGELOG);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "bzip2bin", &flag),
					  RPMVAR_BZIP2BIN);
	Tcl_SetHashValue (Tcl_CreateHashEntry (&jufrpm_vars, "langpatt", &flag),
					  RPMVAR_LANGPATT); 
	
	/* create Tcl commands */
	Tcl_CreateObjCommand (interp, "Juf::RPM::exists", jufrpm_exists,
							  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand (interp, "Juf::RPM::list", jufrpm_list,
						  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand (interp, "Juf::RPM::query", jufrpm_query,
						  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
	Tcl_CreateObjCommand (interp, "Juf::RPM::install", jufrpm_install,
						  (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	/* create variables and tie them to their RPM counterparts */
	entry = Tcl_FirstHashEntry (&jufrpm_vars, &hsrch);
	while (entry != NULL)
		{
		/* variable name */
		Tcl_SetStringObj (buf, "Juf::RPM::", (int) strlen ("Juf::RPM::"));
		name = Tcl_GetHashKey (&jufrpm_vars, entry);
		Tcl_AppendToObj (buf, name, (int) strlen (name));
		/* variable value */
		value = rpmGetVar ((int) Tcl_GetHashValue (entry));
		if (value == NULL)
			value = "";
		/* create variable */
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION == 0)		
		if (Tcl_ObjSetVar2 (interp, buf, NULL,
							Tcl_NewStringObj (value, (int) strlen (value)),
							TCL_LEAVE_ERR_MSG) == NULL)
#else
		if (Tcl_SetObjVar2 (interp, Tcl_GetStringFromObj (buf, NULL), NULL,
							Tcl_NewStringObj (value, (int) strlen (value)),
							TCL_LEAVE_ERR_MSG) == NULL)
#endif			
			return (TCL_ERROR);
		/* tie variable: we use RPM constants to identify them */
		if (Tcl_TraceVar (interp, Tcl_GetStringFromObj (buf, NULL),
						  TCL_TRACE_READS | TCL_TRACE_WRITES,
						  jufrpm_var, Tcl_GetHashValue (entry)) == TCL_ERROR)
			return (TCL_ERROR);
		
		entry = Tcl_NextHashEntry (&hsrch);
		}
	
	/* fill tags hash table */
	Tcl_InitHashTable (&jufrpm_tags, TCL_STRING_KEYS);
	for (i = 0; i < tagmapsize; i++)
		{
		Tcl_SetHashValue
			(Tcl_CreateHashEntry (&jufrpm_tags, tagmap[i].name, &flag),
			 tagmap[i].tag);
		}
	
  
	/* provide JUFRPM as package */
	return (Tcl_PkgProvide (interp, JUFRPM_PKGNAME, JUFRPM_VERSION));
}

