# This is a -*- Tcl -*- library script and belongs to the
# `Jumble Library for Tcl and Friends'.
#
# tpl.tcl --- Template processing functions
#
# Copyright (C) 1997, 1998 Stefan Hornburg
#
# Author: Stefan Hornburg <racke@gundel.han.de>
# Maintainer: Stefan Hornburg <racke@gundel.han.de>
# Version: see end of file
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2, or (at your option) any
# later version.
# 
# This file is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this file; see the file COPYING.  If not, write to the Free
# Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

package require Sequence
package require String 0.0.6

namespace eval Juf {namespace eval Template {}}

# ----------------------------------------------------------------
# PROC: find_macro_delim STRING DELIM START END
#
# Searches for macro delimiter DELIM within STRING and returns
# 1 if successful, 0 otherwise. The variables specified by START
# and END receive start resp. end position of DELIM within STRING.
# ----------------------------------------------------------------

proc Juf::Template::find_macro_delim {string delim start end} {	
	if [regexp -indices $delim $string dlist] {
		upvar $start pos1 $end pos2
		Juf::Sequence::assign $dlist pos1 pos2
		return 1
	}
	return 0
}

proc Juf::Template::find_macro {string start end} {
	if [regexp -indices $start $string v_dummy] {
		set v_pos1 [lindex $v_dummy 0]
		set v_pos2 [lindex $v_dummy 1]
	} else {
		return ""
	}
	set string [string range $string [expr $v_pos2 + 1] end]
	if [regexp -indices $end $string v_dummy] {
		set v_pos3 [expr [lindex $v_dummy 0] + $v_pos2 + 1]
		set v_pos4 [expr [lindex $v_dummy 1] + $v_pos2 + 1]
	} else {
		return ""
	}
	list $v_pos1 $v_pos2 $v_pos3 $v_pos4
}

# ----------------------------------------------------------------
# FUNCTION: schedule START END BETWEEN ARRAYS
#
# Main macro handling routine. START and END are fixed text chunks
# before resp. after the variable part of the macro. BETWEEN is
# the macro itself. ARRAYS contain replacement values.
# Returns result.
# ----------------------------------------------------------------

proc Juf::Template::schedule {start end between arrays} {
	#
	# leading question mark indicates deletion of macro
	# if no value found
	#
	if {[string index $between 0] == "?"} {
		set between [string range $between 1 end]
		set v_preserve 0
	} else {
		set v_preserve 1
	}
	#
	# split macro into variable name and replacement instruction
	#
	set v_list [Juf::String::split $between : 2]
	if {[llength $v_list] == 2} {
		set v_key [lindex $v_list 0]
		set v_instr [lindex $v_list 1]
	} else {
		set v_key $between
		set v_instr "%"
	}
	#
	# array value may be a list
	#
	set v_list [Juf::String::split $v_key # 2]
	if {[llength $v_list] == 2} {
		set v_key [lindex $v_list 0]
		set v_index [lindex $v_list 1]
	} else {
		set v_index -1
	}
	foreach v_aname $arrays {
		upvar 2 $v_aname v_arr
		if [info exists v_arr($v_key)] {
			if {$v_index < 0} {
				#
				# replace with array element
				#
				set v_rpl $v_arr($v_key)
			} else {
				#
				# assume array element is a list
				#
				set v_rpl [lindex $v_arr($v_key) $v_index]
			}
			if [string length $v_rpl] {
				regsub -all % $v_instr $v_rpl v_buf
				return $v_buf
			}
		}
	}
	if $v_preserve {return "$start$between$end"}
}

proc Juf::Template::run {string {arrays ""} {start "@"} {end "@"}} {
	set v_pos 0
	set v_rmd $string
	
	set v_res [Juf::Template::find_macro $v_rmd $start $end]
	while {$v_res != ""} {
		#
		# first add stuff before match to result
		#
		append v_buf [string range $v_rmd 0 [expr [lindex $v_res 0] - 1]]
		#
		# pass macro to scheduler
		#
		append v_buf [Juf::Template::schedule \
				[string range $v_rmd [lindex $v_res 0] [lindex $v_res 1]] \
				[string range $v_rmd [lindex $v_res 2] [lindex $v_res 3]]\
				[string range $v_rmd [expr [lindex $v_res 1] + 1] \
				[expr [lindex $v_res 2] - 1]] \
				$arrays]
		#
		# create string for next macro search
		#
		set v_rmd [string range $v_rmd [expr [lindex $v_res 3] + 1] end]
		set v_res [Juf::Template::find_macro $v_rmd $start $end]
	}
	append v_buf $v_rmd
}

# --------------------------------------------------------------------
# PROC: filter INFD OUTFD [ARRAYS START END]
#
# Reads template from INFD, replaces macro and writes result to OUTFD.
# --------------------------------------------------------------------

proc Juf::Template::filter {infd outfd {arrays ""} {start "@"} {end "@"}} {
	set v_pos 0
	set v_rmd ""
	
	while {[gets $infd line] != -1} {
		if [string length $v_rmd] {
			# pending macro
			if [find_macro_delim $line $end dstart dend] {
				# compose macro
				append v_rmd [lrange $line 0 [incr dstart -1]]
				# process macro and write result
				
			}
		} else {
			# search for macro
			if [find_macro_delim $line $start dstart dend] {
				set v_delim_start [string range $line $dstart $dend]
				if {$dstart > 0} {
					# pass line part before macro to output channel
					puts -nonewline $outfd [string range $line 0 \
							[incr dstart -1]]
				}
				set line [string range $line [incr dend] end]
				# macro completed on this line ?
				if [find_macro_delim $line $end dstart dend2] {
					set v_delim_end [string range $line $dstart $dend2]
					# process macro and pass result to output channel
					puts -nonewline $outfd [schedule $v_delim_start \
							$v_delim_end [string range $line 0 \
							[incr dstart -1]] $arrays]
					# pass line part after macro to output channel
					puts $outfd [string range $line [incr dend2] end]
				} else {
					# record part of multiline macro
					set v_rmd [string range $line [incr dend] end]
				}
			} else {
				# pass line through to output channel
				puts $outfd $line
			}
		}
	}
}

# serves as file version number too
package provide Template 0.0.9
